# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""LibriTTS dataset."""

from __future__ import annotations

import io
import os
import tarfile

import numpy as np
import six
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_URL = "http://www.openslr.org/60"
_DL_URL = "http://www.openslr.org/resources/60/"
_DL_URLS = {
    "dev_clean": _DL_URL + "dev-clean.tar.gz",
    "dev_other": _DL_URL + "dev-other.tar.gz",
    "test_clean": _DL_URL + "test-clean.tar.gz",
    "test_other": _DL_URL + "test-other.tar.gz",
    "train_clean100": _DL_URL + "train-clean-100.tar.gz",
    "train_clean360": _DL_URL + "train-clean-360.tar.gz",
    "train_other500": _DL_URL + "train-other-500.tar.gz",
}


class Builder(tfds.core.BeamBasedBuilder):
  """LibriTTS dataset."""

  VERSION = tfds.core.Version("1.0.1")

  def _info(self):
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            "speech": tfds.features.Audio(file_format="wav", sample_rate=24000),
            "text_original": tfds.features.Text(),
            "text_normalized": tfds.features.Text(),
            "speaker_id": np.int64,
            "chapter_id": np.int64,
            "id": np.str_,
        }),
        supervised_keys=("text_normalized", "speech"),
        homepage=_URL,
        metadata=tfds.core.MetadataDict(
            sample_rate=24000,
        ),
    )

  def _populate_metadata(self, archive_paths):
    # All archives contain the same metadata.
    archive_path = list(archive_paths.values())[0]

    speaker_info = {}
    with tf.io.gfile.GFile(archive_path, "rb") as f:
      tarf = tarfile.open(mode="r:gz", fileobj=f)
      speakers_tsv = tarf.extractfile("LibriTTS/speakers.tsv")
      for n, line in enumerate(speakers_tsv):
        # Skip the first line which is just a header.
        if n == 0:
          continue
        fields = line.decode("utf-8").strip().split("\t")
        if len(fields) == 3:
          # Some lines are missing the final field, so leave it blank.
          fields.append("")
        id_str, gender, subset, name = fields
        speaker_info[int(id_str)] = {
            "gender": gender,
            "subset": subset,
            "name": name,
        }
    self.info.metadata["speakers"] = speaker_info

  def _split_generators(self, dl_manager):
    archives = dl_manager.download(_DL_URLS)
    self._populate_metadata(archives)
    splits = [
        tfds.core.SplitGenerator(name=k, gen_kwargs={"archive_path": v})
        for k, v in archives.items()
    ]
    return splits

  def _build_pcollection(self, pipeline, archive_path):
    """Generates examples as dicts."""
    beam = tfds.core.lazy_imports.apache_beam
    return (
        pipeline
        | beam.Create([archive_path])
        | beam.FlatMap(_extract_libritts_data)
        | beam.Reshuffle()
        | "merge_transcripts_and_audio" >> beam.CombinePerKey(_merge_dicts)
    )


def _generate_transcripts(transcript_csv_file):
  """Generates partial examples from transcript CSV file."""
  for line in transcript_csv_file:
    key, text_original, text_normalized = line.decode("utf-8").split("\t")
    speaker_id, chapter_id = [int(el) for el in key.split("_")[:2]]
    example = {
        "text_normalized": text_normalized,
        "text_original": text_original,
        "speaker_id": speaker_id,
        "chapter_id": chapter_id,
        "id": key,
    }
    yield key, example


def _extract_libritts_data(archive_path):
  """Generate partial audio or transcript examples from a LibriTTS archive."""
  for path, contents in tfds.core.download.extractor.iter_tar(archive_path):
    if path.endswith(".trans.tsv"):
      for key, example in _generate_transcripts(contents):
        yield key, example
    elif path.endswith(".wav"):
      key = six.ensure_text(os.path.splitext(os.path.basename(path))[0])
      memfile = io.BytesIO(contents.read())
      example = {"speech": memfile}
      yield key, example


def _merge_dicts(dicts):
  merged = {}
  for d in dicts:
    merged.update(d)
  return merged



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""tatoeba dataset."""
import os

from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_LANGS = {
    "af": "afr",
    "ar": "ara",
    "bg": "bul",
    "bn": "ben",
    "de": "deu",
    "el": "ell",
    "es": "spa",
    "et": "est",
    "eu": "eus",
    "fa": "pes",
    "fi": "fin",
    "fr": "fra",
    "he": "heb",
    "hi": "hin",
    "hu": "hun",
    "id": "ind",
    "it": "ita",
    "ja": "jpn",
    "jv": "jav",
    "ka": "kat",
    "kk": "kaz",
    "ko": "kor",
    "ml": "mal",
    "mr": "mar",
    "nl": "nld",
    "pt": "por",
    "ru": "rus",
    "sw": "swh",
    "ta": "tam",
    "te": "tel",
    "th": "tha",
    "tl": "tgl",
    "tr": "tur",
    "ur": "urd",
    "vi": "vie",
    "zh": "cmn",
}

_DATA_URLS = "https://raw.githubusercontent.com/facebookresearch/LASER/main/data/tatoeba/v1"


class TatoebaConfig(tfds.core.BuilderConfig):
  """Configuration Class for Tatoeba."""

  def __init__(self, *, language, **kwargs):
    if language not in _LANGS:
      raise ValueError("language must be one of {}".format(list(_LANGS.keys())))

    super(TatoebaConfig, self).__init__(**kwargs)
    self.language = language


class Builder(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for tatoeba dataset."""

  BUILDER_CONFIGS = [
      TatoebaConfig(  # pylint: disable=g-complex-comprehension
          name="tatoeba_" + language,
          language=language,
      )
      for language in _LANGS.keys()
  ]

  VERSION = tfds.core.Version("1.0.0")
  RELEASE_NOTES = {
      "1.0.0": "Initial release.",
  }

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            "source_sentence": tfds.features.Text(),
            "target_sentence": tfds.features.Text(),
            "source_language": tfds.features.Text(),
            "target_language": tfds.features.Text(),
        }),
        supervised_keys=None,
        homepage="http://opus.nlpl.eu/Tatoeba.php",
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    lang = _LANGS[self.builder_config.language]

    tatoeba_source_path = os.path.join(
        _DATA_URLS, "tatoeba.{lang}-eng.{lang}".format(lang=lang)
    )
    tatoeba_eng_path = os.path.join(
        _DATA_URLS, "tatoeba.{}-eng.eng".format(lang)
    )

    archive = dl_manager.download_and_extract({
        "tatoeba_source_data": tatoeba_source_path,
        "tatoeba_eng_data": tatoeba_eng_path,
    })

    return {
        "train": self._generate_examples(
            source_file=archive["tatoeba_source_data"],
            target_file=archive["tatoeba_eng_data"],
        ),
    }

  def _generate_examples(self, source_file, target_file):
    """Yields examples."""
    source_sentences = []
    target_sentences = []
    with tf.io.gfile.GFile(source_file, "rb") as f1:
      for row in f1:
        source_sentences.append(row)
    with tf.io.gfile.GFile(target_file, "rb") as f2:
      for row in f2:
        target_sentences.append(row)
    for i in range(len(source_sentences)):
      yield i, {
          "source_sentence": source_sentences[i],
          "target_sentence": target_sentences[i],
          "source_language": self.builder_config.language,
          "target_language": "en",
      }



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The ImageNet-v2 image classification dataset."""
import os

from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
from tensorflow_datasets.datasets.imagenet2012 import imagenet_common
import tensorflow_datasets.public_api as tfds

# Note: Bump the version if the links change.

_ROOT_URL = 'https://huggingface.co/datasets/vaishaal/ImageNetV2/resolve/main'
_IMAGENET_V2_URLS = {
    'matched-frequency': _ROOT_URL + '/imagenetv2-matched-frequency.tar.gz',
    'threshold-0.7': _ROOT_URL + '/imagenetv2-threshold0.7.tar.gz',
    'topimages': _ROOT_URL + '/imagenetv2-top-images.tar.gz',
}
_TAR_TOPDIR = {
    'matched-frequency': 'imagenetv2-matched-frequency-format-val',
    'threshold-0.7': 'imagenetv2-threshold0.7-format-val',
    'topimages': 'imagenetv2-top-images-format-val',
}


class ImagenetV2Config(tfds.core.BuilderConfig):
  """Configuration specifying the variant to use."""

  def __init__(self, *, variant, **kwargs):
    """The parameters specifying how the dataset will be processed.

    The dataset comes in three different variants. Please refer to the paper
    on more details how they were collected.

    Args:
      variant: One of 'matched-frequency', 'threshold-0.7', or 'topimages'.
      **kwargs: Passed on to the constructor of `BuilderConfig`.
    """
    super(ImagenetV2Config, self).__init__(**kwargs)
    if variant not in _IMAGENET_V2_URLS:
      raise ValueError(
          'Unknown split number {}, must be one of {}'.format(
              variant, list(_IMAGENET_V2_URLS)
          )
      )
    self.variant = variant


def _create_builder_configs():
  for variant in _IMAGENET_V2_URLS:
    yield ImagenetV2Config(
        variant=variant,
        name=variant,
    )


class Builder(tfds.core.GeneratorBasedBuilder):
  """An ImageNet test set recollected by following the original protocol."""

  VERSION = tfds.core.Version('3.0.0')
  SUPPORTED_VERSIONS = [
      tfds.core.Version('2.0.0'),
  ]
  RELEASE_NOTES = {
      '1.0.0': 'Initial version.',
      '2.0.0': 'Files updated.',
      '3.0.0': (
          'Fix file_name, from absolute path to path relative to '
          'data directory, ie: "class_id/filename.jpg".'
      ),
      '3.1.0': 'New URLs for resources from Hugging Face.',
  }
  BUILDER_CONFIGS = list(_create_builder_configs())

  def _info(self):
    names_file = imagenet_common.label_names_file()
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            'image': tfds.features.Image(encoding_format='jpeg'),
            'label': tfds.features.ClassLabel(names_file=names_file),
            'file_name': tfds.features.Text(),
        }),
        # Used if as_supervised=True in builder.as_dataset.
        supervised_keys=('image', 'label'),
        # Homepage of the dataset for documentation
        homepage='https://github.com/modestyachts/ImageNetV2',
    )

  def _split_generators(self, dl_manager):
    """Returns a SplitGenerator for the test set."""
    variant_url = _IMAGENET_V2_URLS[self.builder_config.variant]
    imagenet_v2_root = os.path.join(
        dl_manager.download_and_extract(variant_url),
        _TAR_TOPDIR[self.builder_config.variant],
    )
    return [
        tfds.core.SplitGenerator(
            # The dataset provides only a test split.
            name=tfds.Split.TEST,
            # These kwargs will be passed to _generate_examples
            gen_kwargs={'imagenet_v2_root': imagenet_v2_root},
        ),
    ]

  def _generate_examples(self, imagenet_v2_root):
    """Yields the examples."""
    # The directory structure is `class_number/filename_number.jpg`, where
    # class_number is in [0, 1000) and filename_number in [0, 10).
    for class_id in tf.io.gfile.listdir(imagenet_v2_root):
      class_dir = os.path.join(imagenet_v2_root, class_id)
      for image_filename in tf.io.gfile.listdir(class_dir):
        image_path = os.path.join(class_dir, image_filename)
        features = {
            'image': image_path,
            'label': int(class_id),
            'file_name': os.path.join(class_id, image_filename),
        }
        yield f'{class_id}_{image_filename}', features



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions to convert from other formats to TFDS conventions."""

from collections.abc import Mapping, Sequence
import datetime
from typing import Any

from etils import epath
import numpy as np
from tensorflow_datasets.core import features as feature_lib
from tensorflow_datasets.core import lazy_imports_lib
from tensorflow_datasets.core.utils import dtype_utils
from tensorflow_datasets.core.utils import py_utils

_DEFAULT_IMG = b'\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x01\x00\x00\x00\x01\x08\x02\x00\x00\x00\x90wS\xde\x00\x00\x00\x0cIDATx\x9cc```\x00\x00\x00\x04\x00\x01\xf6\x178U\x00\x00\x00\x00IEND\xaeB`\x82'


def to_tfds_name(name: str) -> str:
  """Converts a name to a TFDS compatible dataset name.

  Huggingface names can contain characters that are not supported in
  TFDS. For example, in Huggingface a dataset name like `a/b` is supported,
  while in TFDS `b` would be parsed as the config.

  Examples:
  - `name='codeparrot/github-code'` becomes `codeparrot__github_code`.

  Args:
    name: A name to be converted to a TFDS compatible name.

  Returns:
    The TFDS compatible dataset name (dataset names, config names and split
    names).
  """
  name = name.lower().replace('/', '__')
  return py_utils.make_valid_name(name)


def _get_default_value(
    feature: feature_lib.FeatureConnector,
) -> Mapping[str, Any] | Sequence[Any] | bytes | int | float | bool:
  """Returns the default value for a feature.

  Non-TFDS features can be loose as far as typing is concerned. For example,
  HuggingFace accepts None values. As long as `tfds.features.Optional` does not
  exist, we default to a constant default value.

  For int and float, we do not return 0 or -1, but rather -inf, as 0 or -1 can
  be contained in the values of the dataset. In practice, you can compare your
  value to:

  ```
  np.iinfo(np.int32).min  # for integers
  np.finfo(np.float32).min  # for floats
  ...
  ```

  For None images, we set a default value which corresponds to a PNG of 1px,
  black.

  Args:
    feature: The TFDS feature from which we want the default value.

  Raises:
    TypeError: If couldn't recognize feature dtype.
  """
  match feature:
    case feature_lib.FeaturesDict():
      return {
          name: _get_default_value(inner_feature)
          for name, inner_feature in feature.items()
      }
    case feature_lib.Sequence():
      match feature.feature:
        case feature_lib.FeaturesDict():
          return {feature_name: [] for feature_name in feature.feature.keys()}
        case _:
          return []
    case feature_lib.Image():
      # Return an empty PNG image of 1x1 pixel, black.
      return _DEFAULT_IMG
    case _:
      if dtype_utils.is_string(feature.np_dtype):
        return b''
      elif dtype_utils.is_integer(feature.np_dtype):
        return np.iinfo(feature.np_dtype).min
      elif dtype_utils.is_floating(feature.np_dtype):
        return np.finfo(feature.np_dtype).min
      elif dtype_utils.is_bool(feature.np_dtype):
        return False
      else:
        raise TypeError(f'Could not recognize the dtype of {feature}')


def to_tfds_value(value: Any, feature: feature_lib.FeatureConnector) -> Any:
  """Converts a value to a TFDS compatible value.

  Args:
    value: The value to be converted to follow TFDS conventions.
    feature: The TFDS feature for which we want the compatible value.

  Returns:
    The TFDS compatible value.

  Raises:
    TypeError: If couldn't recognize the given feature type.
  """
  match value:
    case None:
      return _get_default_value(feature)
    case datetime.datetime():
      return int(value.timestamp())

  match feature:
    case feature_lib.ClassLabel() | feature_lib.Scalar():
      return value
    case feature_lib.Video():
      match value:
        case dict():
          if 'path' in value and value['path']:
            return value['path']
          elif 'bytes' in value and value['bytes']:
            return value['bytes']
          else:
            raise ValueError(
                'Dictionary-like video features must have either a `path` or'
                ' `bytes` key.'
            )
    case feature_lib.FeaturesDict():
      return {
          name: to_tfds_value(value.get(name), inner_feature)
          for name, inner_feature in feature.items()
      }
    case feature_lib.Sequence():
      match value:
        case dict():
          # Should be a dict of lists:
          return {
              name: [
                  to_tfds_value(inner_hf_value, inner_feature)
                  for inner_hf_value in value.get(name)
              ]
              for name, inner_feature in feature.feature.items()
          }
        case list():
          return [
              to_tfds_value(inner_hf_value, feature.feature)
              for inner_hf_value in value
          ]
        case _:
          return [value]
    case feature_lib.Audio():
      if (array := value.get('array')) is not None:
        # Hugging Face uses floats, TFDS uses integers.
        # Here we convert the float in [-1, 1] range into signed int32
        # range [-2**32, 2**32-1]. Nevertheless, the mantissa size of
        # float32 is 23 bits, therefore the maximum bit depth possible is 23.
        dtype = feature.dtype
        return (array * np.iinfo(dtype).max).astype(dtype=dtype)
      elif (path := value.get('path')) and (path := epath.Path(path)).exists():
        return path
    case feature_lib.Image():
      value: lazy_imports_lib.lazy_imports.PIL_Image.Image
      # Ensure RGB format for PNG encoding.
      return value.convert('RGB')
    case feature_lib.Tensor():
      if isinstance(value, float):
        # In some cases, for example when loading jsonline files using pandas,
        # empty non-float values, such as strings, are converted to float nan.
        # We spot those occurrences as the feature.np_dtype is not float.
        if np.isnan(value) and not dtype_utils.is_floating(feature.np_dtype):
          return _get_default_value(feature)
      return value

  raise TypeError(
      f'Conversion of value {value} to feature {feature} is not supported.'
  )



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""QA4MRE (CLEF 2011/2012/2013): a reading comprehension dataset."""

import os
import xml.etree.ElementTree as ET

from absl import logging
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

# pylint: disable=anomalous-backslash-in-string

_BASE_URL = 'http://nlp.uned.es/clef-qa/repository/js/scripts/downloadFile.php?file=/var/www/html/nlp/clef-qa/repository/resources/QA4MRE/'

PATHS = {
    '2011': {
        '_TRACKS': 'main',
        '_PATH_TMPL_MAIN_GS': (
            '2011/Training_Data/Goldstandard/QA4MRE-2011-{}_GS.xml'
        ),
        '_LANGUAGES_MAIN': ('DE', 'EN', 'ES', 'IT', 'RO'),
    },
    '2012': {
        '_TRACKS': ('main', 'alzheimers'),
        '_PATH_TMPL_MAIN_GS': '2012/Main_Task/Training_Data/Goldstandard/Used_in_Evaluation/QA4MRE-2012-{}_GS.xml',
        '_LANGUAGES_MAIN': ('AR', 'BG', 'DE', 'EN', 'ES', 'IT', 'RO'),
        '_PATH_ALZHEIMER': '2012/Pilot_Tasks/Biomedical_About_Alzheimer/Training_Data/Goldstandard/QA4MRE-2012_BIOMEDICAL_GS.xml',
    },
    '2013': {
        '_TRACKS': ('main', 'alzheimers', 'entrance_exam'),
        '_PATH_TMPL_MAIN_GS': (
            '2013/Main_Task/Training_Data/Goldstandard/QA4MRE-2013-{}_GS.xml'
        ),
        '_LANGUAGES_MAIN': ('AR', 'BG', 'EN', 'ES', 'RO'),
        '_PATH_ALZHEIMER': '2013/Biomedical_About_Alzheimer/Training_Data/Goldstandard/QA4MRE-2013_BIO_GS-RUN.xml',
        '_PATH_ENTRANCE_EXAM': '2013/Entrance_Exams/Training_Data/Goldstandard/qa4mre-exam-test-withanswer.xml',
    },
}


def _get_question(
    topic_id, topic_name, test_id, document_id, document_str, question
):
  """Gets instance ID and features for every question.

  Args:
    topic_id: string
    topic_name: string
    test_id: string
    document_id: string
    document_str: string
    question: XML element for question

  Returns:
    id_: string. Unique ID for instance.
    feats: dict of instance features
  """

  question_id = question.attrib['q_id']
  for q_text in question.iter('q_str'):
    question_str = q_text.text
  possible_answers = list()
  for answer in question.iter('answer'):
    answer_id = answer.attrib['a_id']
    answer_str = answer.text
    possible_answers.append({'answer_id': answer_id, 'answer_str': answer_str})
    if 'correct' in answer.attrib:
      correct_answer_id = answer_id
      correct_answer_str = answer_str

  id_ = '_'.join([topic_id, topic_name, test_id, question_id])
  logging.info('ID: %s', id_)

  feats = {
      'topic_id': topic_id,
      'topic_name': topic_name,
      'test_id': test_id,
      'document_id': document_id,
      'document_str': document_str,
      'question_id': question_id,
      'question_str': question_str,
      'answer_options': possible_answers,
      'correct_answer_id': correct_answer_id,
      'correct_answer_str': correct_answer_str,
  }

  return id_, feats


class Qa4mreConfig(tfds.core.BuilderConfig):
  """BuilderConfig for Qa4mre."""

  def __init__(self, *, year, track='main', language='EN', **kwargs):
    """BuilderConfig for Qa4Mre.

    Args:
      year: string, year of dataset
      track: string, the task track from PATHS[year]['_TRACKS'].
      language: string, Acronym for language in the main task.
      **kwargs: keyword arguments forwarded to super.
    """
    if track.lower() not in PATHS[year]['_TRACKS']:
      raise ValueError(
          'Incorrect track. Track should be one of the following: ',
          PATHS[year]['_TRACKS'],
      )

    if track.lower() != 'main' and language.upper() != 'EN':
      logging.warning(
          'Only English documents available for pilot '
          'tracks. Setting English by default.'
      )
      language = 'EN'

    if (
        track.lower() == 'main'
        and language.upper() not in PATHS[year]['_LANGUAGES_MAIN']
    ):
      raise ValueError(
          'Incorrect language for the main track. Correct options: ',
          PATHS[year]['_LANGUAGES_MAIN'],
      )

    self.year = year
    self.track = track.lower()
    self.lang = language.upper()

    name = self.year + '.' + self.track + '.' + self.lang

    description = (
        'This configuration includes the {} track for {} language in {} year.'
    ).format(self.track, self.lang, self.year)

    super(Qa4mreConfig, self).__init__(
        name=name,
        description=description,
        version=tfds.core.Version('0.1.0'),
        **kwargs,
    )


class Builder(tfds.core.GeneratorBasedBuilder):
  """QA4MRE dataset from CLEF shared tasks 2011, 2012, 2013."""

  BUILDER_CONFIGS = [
      Qa4mreConfig(
          year='2011', track='main', language='DE'
      ),  # 2011 Main track German (2011.main.DE)
      Qa4mreConfig(
          year='2011', track='main', language='EN'
      ),  # 2011 Main track English (2011.main.EN)
      Qa4mreConfig(
          year='2011', track='main', language='ES'
      ),  # 2011 Main track Spanish (2011.main.ES)
      Qa4mreConfig(
          year='2011', track='main', language='IT'
      ),  # 2011 Main track Italian (2011.main.IT)
      Qa4mreConfig(
          year='2011', track='main', language='RO'
      ),  # 2011 Main track Romanian (2011.main.RO)
      Qa4mreConfig(
          year='2012', track='main', language='AR'
      ),  # 2012 Main track Arabic (2012.main.AR)
      Qa4mreConfig(
          year='2012', track='main', language='BG'
      ),  # 2012 Main track Bulgarian (2012.main.BG)
      Qa4mreConfig(
          year='2012', track='main', language='DE'
      ),  # 2012 Main track German (2012.main.DE)
      Qa4mreConfig(
          year='2012', track='main', language='EN'
      ),  # 2012 Main track English (2012.main.EN)
      Qa4mreConfig(
          year='2012', track='main', language='ES'
      ),  # 2012 Main track Spanish (2012.main.ES)
      Qa4mreConfig(
          year='2012', track='main', language='IT'
      ),  # 2012 Main track Italian (2012.main.IT)
      Qa4mreConfig(
          year='2012', track='main', language='RO'
      ),  # 2012 Main track Romanian (2012.main.RO)
      Qa4mreConfig(
          year='2012', track='alzheimers', language='EN'
      ),  # (2012.alzheimers.EN)
      Qa4mreConfig(
          year='2013', track='main', language='AR'
      ),  # 2013 Main track Arabic (2013.main.AR)
      Qa4mreConfig(
          year='2013', track='main', language='BG'
      ),  # 2013 Main track Bulgarian (2013.main.BG)
      Qa4mreConfig(
          year='2013', track='main', language='EN'
      ),  # 2013 Main track English (2013.main.EN)
      Qa4mreConfig(
          year='2013', track='main', language='ES'
      ),  # 2013 Main track Spanish (2013.main.ES)
      Qa4mreConfig(
          year='2013', track='main', language='RO'
      ),  # 2013 Main track Romanian (2013.main.RO)
      Qa4mreConfig(
          year='2013', track='alzheimers', language='EN'
      ),  # (2013.alzheimers.EN)
      Qa4mreConfig(
          year='2013', track='entrance_exam', language='EN'
      ),  # (2013.entrance_exam.EN)
  ]

  def _info(self):
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            'topic_id': tfds.features.Text(),
            'topic_name': tfds.features.Text(),
            'test_id': tfds.features.Text(),
            'document_id': tfds.features.Text(),
            'document_str': tfds.features.Text(),
            'question_id': tfds.features.Text(),
            'question_str': tfds.features.Text(),
            'answer_options': tfds.features.Sequence({
                'answer_id': tfds.features.Text(),
                'answer_str': tfds.features.Text(),
            }),
            'correct_answer_id': tfds.features.Text(),
            'correct_answer_str': tfds.features.Text(),
        }),
        # No default supervised keys because both passage and question are used
        # to determine the correct answer.
        supervised_keys=None,
        homepage='http://nlp.uned.es/clef-qa/repository/pastCampaigns.php',
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""
    cfg = self.builder_config
    download_urls = dict()

    if cfg.track == 'main':
      download_urls['{}.main.{}'.format(cfg.year, cfg.lang)] = os.path.join(
          _BASE_URL, PATHS[cfg.year]['_PATH_TMPL_MAIN_GS'].format(cfg.lang)
      )  # pytype: disable=attribute-error

    if cfg.year in ['2012', '2013'] and cfg.track == 'alzheimers':
      download_urls['{}.alzheimers.EN'.format(cfg.year)] = os.path.join(
          _BASE_URL, PATHS[cfg.year]['_PATH_ALZHEIMER']
      )

    if cfg.year == '2013' and cfg.track == 'entrance_exam':
      download_urls['2013.entrance_exam.EN'] = os.path.join(
          _BASE_URL, PATHS[cfg.year]['_PATH_ENTRANCE_EXAM']
      )

    downloaded_files = dl_manager.download_and_extract(download_urls)

    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs={
                'filepath': downloaded_files[
                    '{}.{}.{}'.format(cfg.year, cfg.track, cfg.lang)
                ]
            },
        )
    ]

  def _generate_examples(self, filepath):
    """Yields examples."""
    with tf.io.gfile.GFile(filepath, 'rb') as f:
      tree = ET.parse(f)
      root = tree.getroot()  # test-set
      for topic in root:
        topic_id = topic.attrib['t_id']
        topic_name = topic.attrib['t_name']
        for test in topic:
          test_id = test.attrib['r_id']
          for document in test.iter('doc'):
            document_id = document.attrib['d_id']
            document_str = document.text
          for question in test.iter('q'):
            yield _get_question(
                topic_id,
                topic_name,
                test_id,
                document_id,
                document_str,
                question,
            )



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for the assin2 dataset."""

import dataclasses
from typing import List, Optional
from xml.etree import cElementTree as ET

_VALID_ENTAILMENT = ('Entailment', 'None')


@dataclasses.dataclass
class Pair:
  text: str
  hypothesis: str
  id: int
  entailment: str
  similarity: float


def get_element_text(xml_pair: ET.Element, tag: str) -> Optional[str]:
  """Returns the text associated with a given tag."""
  element = xml_pair.find(tag)
  assert element is not None, f'The tag "{tag}" was not found.'
  return element.text


def parse_xml_string(xml_str: str) -> List[Pair]:
  """Process ASSIN2 xml files content.

  Modified from https://github.com/erickrf/assin.

  Args:
    xml_str: string with XML contents

  Returns:
    list of Pair objects
  """
  pairs = []
  root = ET.fromstring(xml_str)

  for xml_pair in root.iter('pair'):
    # Get fields.
    text = get_element_text(xml_pair, 't')
    hypothesis = get_element_text(xml_pair, 'h')

    attribs = dict(xml_pair.items())
    id_ = int(attribs['id'])
    entailment = attribs['entailment']
    similarity = float(attribs['similarity'])

    # Assert valid fields.
    if entailment not in _VALID_ENTAILMENT:
      raise ValueError(
          f'Entailment should be in {_VALID_ENTAILMENT}, got {entailment}.'
      )
    pairs.append(
        Pair(
            text=text,
            hypothesis=hypothesis,
            id=id_,
            entailment=entailment,
            similarity=similarity,
        )
    )
  return pairs



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Healthy and unhealthy plant leaves dataset."""

import os

import tensorflow_datasets.public_api as tfds

# File name prefix to label mapping.
_LABEL_MAPPING = [
    ("0001", "Mango (P0) healthy"),
    ("0002", "Arjun (P1) healthy"),
    ("0003", "Alstonia Scholaris (P2) healthy"),
    ("0004", "Gauva (P3) healthy"),
    ("0005", "Jamun (P5) healthy"),
    ("0006", "Jatropha (P6) healthy"),
    ("0007", "Pongamia Pinnata (P7) healthy"),
    ("0008", "Basil (P8) healthy"),
    ("0009", "Pomegranate (P9) healthy"),
    ("0010", "Lemon (P10) healthy"),
    ("0011", "Chinar (P11) healthy"),
    ("0012", "Mango (P0) diseased"),
    ("0013", "Arjun (P1) diseased"),
    ("0014", "Alstonia Scholaris (P2) diseased"),
    ("0015", "Gauva (P3) diseased"),
    ("0016", "Bael (P4) diseased"),
    ("0017", "Jamun (P5) diseased"),
    ("0018", "Jatropha (P6) diseased"),
    ("0019", "Pongamia Pinnata (P7) diseased"),
    ("0020", "Pomegranate (P9) diseased"),
    ("0021", "Lemon (P10) diseased"),
    ("0022", "Chinar (P11) diseased"),
]

_DOWNLOAD_URL = "https://prod-dcd-datasets-cache-zipfiles.s3.eu-west-1.amazonaws.com/hb74ynkjcn-1.zip"


class DownloadRetryLimitReachedError(Exception):
  pass


class Builder(tfds.core.GeneratorBasedBuilder):
  """Healthy and unhealthy plant leaves dataset."""

  VERSION = tfds.core.Version("0.1.1")
  RELEASE_NOTES = {
      "0.1.0": "Initial release.",
      "0.1.1": "Fix checksum error.",
  }

  def _info(self):
    labels = list(zip(*_LABEL_MAPPING))[1]
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            "image": tfds.features.Image(),
            "image/filename": tfds.features.Text(),
            "label": tfds.features.ClassLabel(names=labels),
        }),
        supervised_keys=("image", "label"),
        homepage="https://data.mendeley.com/datasets/hb74ynkjcn/1",
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""
    # Batch download for this dataset is broken, therefore images have to be
    # downloaded independently from a list of urls.
    archive_path = dl_manager.download(_DOWNLOAD_URL)
    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs={"archive": dl_manager.iter_archive(archive_path)},
        )
    ]

  def _generate_examples(self, archive):
    """Yields examples."""
    for filename, fobj in archive:
      if not filename.endswith(".JPG"):
        continue
      fname_split = filename.split(os.path.sep)
      label = fname_split[-3] + " " + fname_split[-2]
      record = {
          "image": fobj,
          "image/filename": fname_split[-1],
          "label": label,
      }
      yield fname_split[-1], record



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for corrupted_imagenet."""

from tensorflow_datasets import testing
from tensorflow_datasets.datasets.imagenet2012_corrupted import imagenet2012_corrupted_dataset_builder


class Imagenet2012CorruptedTest(testing.DatasetBuilderTestCase):
  BUILDER_CONFIG_NAMES_TO_TEST = [  # pylint: disable=g-long-ternary
      "gaussian_noise_1",
      "shot_noise_2",
      "impulse_noise_3",
      "defocus_blur_4",
      "glass_blur_5",
      "motion_blur_1",
      "zoom_blur_2",
      "snow_3",
      "frost_4",
      "fog_5",
      "brightness_1",
      "contrast_2",
      "elastic_transform_3",
      "pixelate_4",
      "jpeg_compression_5",
      "gaussian_blur_1",
      "saturate_2",
      "spatter_3",
      "speckle_noise_4",
  ]

  DATASET_CLASS = imagenet2012_corrupted_dataset_builder.Builder
  SPLITS = {  # Expected number of examples on the train/validation splits.
      "validation": 10,
  }
  DL_EXTRACT_RESULT = [
      "ILSVRC2012_img_train.tar",
      "ILSVRC2012_img_val.tar",
  ]
  DL_DOWNLOAD_RESULT = [
      "frost1.png",
      "frost2.png",
      "frost3.png",
      "frost4.jpg",
      "frost5.jpg",
      "frost6.jpg",
  ]


if __name__ == "__main__":
  testing.test_main()



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Natural Questions Open: A Benchmark for Open Domain Question Answering."""

from __future__ import annotations

import json

from etils import epath
import numpy as np
import tensorflow_datasets.public_api as tfds

_DOWNLOAD_URL_FMT = 'https://raw.githubusercontent.com/google-research-datasets/natural-questions/a7a113c9fdc2d9986d624bd43b8a18d6d5779eaa/nq_open/NQ-open.{split}.jsonl'

_URL = 'https://github.com/google-research-datasets/natural-questions/tree/master/nq_open'


class Builder(tfds.core.GeneratorBasedBuilder):
  """Natural Questions Open: A Benchmark for Open Domain Question Answering."""

  VERSION = tfds.core.Version('1.0.0')

  def _info(self):
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict(
            {'question': np.str_, 'answer': tfds.features.Sequence(np.str_)}
        ),
        supervised_keys=None,
        homepage=_URL,
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""

    file_paths = dl_manager.download({
        'train': _DOWNLOAD_URL_FMT.format(split='train'),
        'validation': _DOWNLOAD_URL_FMT.format(split='dev'),
    })
    return [
        tfds.core.SplitGenerator(
            name=split, gen_kwargs={'file_path': file_path}
        )
        for split, file_path in file_paths.items()
    ]

  def _generate_examples(self, file_path):
    """Parses split file and yields examples."""

    with epath.Path(file_path).open() as f:
      for i, line in enumerate(f):
        yield i, json.loads(line)



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Download manager interface."""

from __future__ import annotations

from collections.abc import Iterator
import concurrent.futures
import dataclasses
import functools
import typing
from typing import Any
import uuid

from absl import logging
from etils import epath
from etils import epy
from tensorflow_datasets.core.utils.lazy_imports_utils import tree

with epy.lazy_imports():
  # pylint: disable=g-import-not-at-top
  import promise

  from tensorflow_datasets.core import utils
  from tensorflow_datasets.core.download import checksums
  from tensorflow_datasets.core.download import downloader
  from tensorflow_datasets.core.download import extractor
  from tensorflow_datasets.core.download import kaggle
  from tensorflow_datasets.core.download import resource as resource_lib
  from tensorflow_datasets.core.download import util
  from tensorflow_datasets.core.utils import shard_utils
  from tensorflow_datasets.core.utils import type_utils
  # pylint: enable=g-import-not-at-top

# pylint: disable=logging-fstring-interpolation

Tree = type_utils.Tree

Url = str | resource_lib.Resource
ExtractPath = epath.PathLike | resource_lib.Resource


def get_downloader(*args: Any, **kwargs: Any) -> downloader._Downloader:
  return downloader.get_downloader(*args, **kwargs)


class NonMatchingChecksumError(Exception):
  """The downloaded file doesn't have expected checksum."""


# Even if `DownloadConfig` is immutable inside TFDS, we do not set `frozen=True`
# to allow user to set:
# dl_config = tfds.download.DownloadConfig()
# dl_config.beam_runner = ...
@dataclasses.dataclass(eq=False)
class DownloadConfig:
  """Configuration for `tfds.core.DatasetBuilder.download_and_prepare`.

  Attributes:
    extract_dir: `str`, directory where extracted files are stored. Defaults to
      "<download_dir>/extracted".
    manual_dir: `str`, read-only directory where manually downloaded/extracted
      data is stored. Defaults to `<download_dir>/manual`.
    download_mode: `tfds.GenerateMode`, how to deal with downloads or data that
      already exists. Defaults to `REUSE_DATASET_IF_EXISTS`, which will reuse
      both downloads and data if it already exists.
    compute_stats: `tfds.download.ComputeStats`, whether to compute statistics
      over the generated data. Defaults to `AUTO`.
    max_examples_per_split: `int`, optional max number of examples to write into
      each split (used for testing). If set to 0, only execute the
      `_split_generators` (download the original data), but skip
      `_generator_examples`.
    register_checksums: `bool`, defaults to False. If True, checksum of
      downloaded files are recorded.
    force_checksums_validation: `bool`, defaults to False. If True, raises an
      error if an URL do not have checksums.
    beam_runner: Runner to pass to `beam.Pipeline`, only used for datasets based
      on Beam for the generation.
    beam_options: `PipelineOptions` to pass to `beam.Pipeline`, only used for
      datasets based on Beam for the generation.
    try_download_gcs: `bool`, defaults to True. If True, prepared dataset will
      be downloaded from GCS, when available. If False, dataset will be
      downloaded and prepared from scratch.
    verify_ssl: `bool`, defaults to True. If True, will verify certificate when
      downloading dataset.
    override_max_simultaneous_downloads: `int`, optional max number of
      simultaneous downloads. If set, it will override dataset builder and
      downloader default values.
    num_shards: optional number of shards that should be created. If `None`,
      then the number of shards is computed based on the total size of the
      dataset and the min and max shard size.
    min_shard_size: optional minimum shard size in bytes. If `None`, 64 MB is
      used.
    max_shard_size: optional maximum shard size in bytes. If `None`, 1 GiB is
      used.
    ignore_duplicates: whether to ignore duplicated examples with the same key.
      If there are multiple examples with the same key, the first one is kept.
    nondeterministic_order: If True, it will not assure deterministic ordering
      when writing' examples to disk in the case of beam datasets. This might
      result in quicker dataset preparation.
  """

  extract_dir: epath.PathLike | None = None
  manual_dir: epath.PathLike | None = None
  download_mode: util.GenerateMode = util.GenerateMode.REUSE_DATASET_IF_EXISTS
  compute_stats: util.ComputeStatsMode = util.ComputeStatsMode.SKIP
  max_examples_per_split: int | None = None
  register_checksums: bool = False
  force_checksums_validation: bool = False
  beam_runner: Any | None = None
  beam_options: Any | None = None
  try_download_gcs: bool = True
  verify_ssl: bool = True
  override_max_simultaneous_downloads: int | None = None
  num_shards: int | None = None
  min_shard_size: int = shard_utils.DEFAULT_MIN_SHARD_SIZE
  max_shard_size: int = shard_utils.DEFAULT_MAX_SHARD_SIZE
  ignore_duplicates: bool = False
  nondeterministic_order: bool = False

  def get_shard_config(self) -> shard_utils.ShardConfig:
    return shard_utils.ShardConfig(
        num_shards=self.num_shards,
        min_shard_size=self.min_shard_size,
        max_shard_size=self.max_shard_size,
    )

  def replace(self, **kwargs: Any) -> DownloadConfig:
    """Returns a copy with updated attributes."""
    return dataclasses.replace(self, **kwargs)


class DownloadManager:
  """Manages the download and extraction of files, as well as caching.

  Downloaded files are cached under `download_dir`. The file name of downloaded
   files follows pattern "{sanitized_url}{content_checksum}.{ext}". Eg:
   'cs.toronto.edu_kriz_cifar-100-pythonJDF[...]I.tar.gz'.

  While a file is being downloaded, it is placed into a directory following a
  similar but different pattern:
  "{sanitized_url}{url_checksum}.tmp.{uuid}".

  When a file is downloaded, a "{fname}.INFO.json" file is created next to it.
  This INFO file contains the following information:
  {"dataset_names": ["name1", "name2"],
   "urls": ["http://url.of/downloaded_file"]}

  Extracted files/dirs are stored under `extract_dir`. The file name or
  directory name is the same as the original name, prefixed with the extraction
  method. E.g.
   "{extract_dir}/TAR_GZ.cs.toronto.edu_kriz_cifar-100-pythonJDF[...]I.tar.gz".

  The function members accept either plain value, or values wrapped into list
  or dict. Giving a data structure will parallelize the downloads.

  Example of usage:

  ```
  # Sequential download: str -> str
  train_dir = dl_manager.download_and_extract('https://abc.org/train.tar.gz')
  test_dir = dl_manager.download_and_extract('https://abc.org/test.tar.gz')

  # Parallel download: list -> list
  image_files = dl_manager.download(
      ['https://a.org/1.jpg', 'https://a.org/2.jpg', ...])

  # Parallel download: dict -> dict
  data_dirs = dl_manager.download_and_extract({
     'train': 'https://abc.org/train.zip',
     'test': 'https://abc.org/test.zip',
  })
  data_dirs['train']
  data_dirs['test']
  ```

  For more customization on the download/extraction (ex: passwords, output_name,
  ...), you can pass a `tfds.download.Resource` as argument.
  """

  def __init__(
      self,
      *,
      download_dir: epath.PathLike,
      extract_dir: epath.PathLike | None = None,
      manual_dir: epath.PathLike | None = None,
      manual_dir_instructions: str | None = None,
      url_infos: dict[str, checksums.UrlInfo] | None = None,
      dataset_name: str | None = None,
      force_download: bool = False,
      force_extraction: bool = False,
      force_checksums_validation: bool = False,
      register_checksums: bool = False,
      register_checksums_path: epath.PathLike | None = None,
      verify_ssl: bool = True,
      max_simultaneous_downloads: int | None = None,
  ):
    """Download manager constructor.

    Args:
      download_dir: Path to directory where downloads are stored.
      extract_dir: Path to directory where artifacts are extracted.
      manual_dir: Path to manually downloaded/extracted data directory.
      manual_dir_instructions: Human readable instructions on how to prepare
        contents of the manual_dir for this dataset.
      url_infos: Urls info for the checksums.
      dataset_name: Name of dataset this instance will be used for. If provided,
        downloads will contain which datasets they were used for.
      force_download: If True, always [re]download.
      force_extraction: If True, always [re]extract.
      force_checksums_validation: If True, raises an error if an URL do not have
        checksums.
      register_checksums: If True, dl checksums aren't checked, but stored into
        file.
      register_checksums_path: Path were to save checksums. Should be set if
        register_checksums is True.
      verify_ssl: `bool`, defaults to True. If True, will verify certificate
        when downloading dataset.
      max_simultaneous_downloads: `int`, optional max number of simultaneous
        downloads.

    Raises:
      FileNotFoundError: Raised if the register_checksums_path does not exist.
    """
    if register_checksums:
      if not register_checksums_path:
        raise ValueError(
            'When register_checksums=True, register_checksums_path should be'
            ' set.'
        )
      register_checksums_path = epath.Path(register_checksums_path)

    download_dir = epath.Path(download_dir).expanduser()
    if extract_dir:
      extract_dir = epath.Path(extract_dir).expanduser()
    else:
      extract_dir = download_dir / 'extracted'
    if manual_dir is not None:
      manual_dir = epath.Path(manual_dir).expanduser()

    self._download_dir: epath.Path = download_dir
    self._extract_dir: epath.Path = extract_dir
    self._manual_dir: epath.Path | None = manual_dir
    self._manual_dir_instructions = utils.dedent(manual_dir_instructions)
    self._download_dir.mkdir(parents=True, exist_ok=True)
    self._extract_dir.mkdir(parents=True, exist_ok=True)

    self._force_download = force_download
    self._force_extraction = force_extraction
    self._force_checksums_validation = force_checksums_validation
    self._register_checksums = register_checksums
    self._register_checksums_path = register_checksums_path
    self._verify_ssl = verify_ssl
    self._max_simultaneous_downloads = max_simultaneous_downloads
    self._dataset_name = dataset_name

    # All known URLs: {url: UrlInfo(size=, checksum=)}
    self._url_infos = checksums.get_all_url_infos()
    if url_infos is not None:
      self._url_infos.update(url_infos)

    # To record what is being used: {url: UrlInfo(size, checksum, filename)}
    self._recorded_url_infos: dict[str, checksums.UrlInfo] = {}
    # These attributes are lazy-initialized since they must be cleared when this
    # object is pickled for Beam. They are then recreated on each worker.
    self.__downloader = None
    self.__extractor = None
    # Executor to avoid blocking other download/extractions when running I/O
    # operations (reading/renaming download file).
    # Only use a single thread as the read/ops are locked by the
    # `build_synchronize_decorator`.
    # Note: This thread is in additions of the download and extraction
    # executors threads.
    self._executor = concurrent.futures.ThreadPoolExecutor(1)

  def __getstate__(self):
    """Remove un-pickleable attributes and return the state."""
    if self._register_checksums:
      # Currently, checksums registration from Beam not supported.
      raise NotImplementedError(
          '`register_checksums` must be disabled in a parallelized '
          'DownloadManager. Please open a PR if you would like this feature.'
      )
    state = self.__dict__.copy()
    state['_DownloadManager__downloader'] = None
    state['_DownloadManager__extractor'] = None
    state['_executor'] = None
    return state

  @property
  def _downloader(self) -> downloader._Downloader:
    if not self.__downloader:
      self.__downloader = get_downloader(
          max_simultaneous_downloads=self._max_simultaneous_downloads
      )
    return self.__downloader

  @property
  def _extractor(self) -> extractor._Extractor:
    if not self.__extractor:
      self.__extractor = extractor.get_extractor()
    return self.__extractor

  @property
  def downloaded_size(self) -> int:
    """Returns the total size of downloaded files."""
    return sum(url_info.size for url_info in self._recorded_url_infos.values())

  def _get_dl_path(
      self,
      resource: resource_lib.Resource,
      checksum: str | None = None,
      legacy_mode: bool = False,
  ) -> epath.Path:
    """Returns the path where the resource should be downloaded.

    Args:
      resource: The resource to download.
      checksum: The checksum of the resource.
      legacy_mode: If True, returns path in the legacy format without dataset
        name in the path.
    """
    download_dir = self._download_dir
    if not legacy_mode and self._dataset_name:
      download_dir /= self._dataset_name
    download_dir /= resource.relative_download_dir
    return download_dir / resource_lib.get_dl_fname(resource.url, checksum)

  @property
  def register_checksums(self):
    """Returns whether checksums are being computed and recorded to file."""
    return self._register_checksums

  @utils.build_synchronize_decorator()
  def _record_url_infos(self):
    """Store in file when recorded size/checksum of downloaded files."""
    checksums.save_url_infos(
        self._register_checksums_path,
        self._recorded_url_infos,
    )

  def _get_manually_downloaded_path(
      self, expected_url_info: checksums.UrlInfo | None
  ) -> epath.Path | None:
    """Checks if file is already downloaded in manual_dir."""
    if not self._manual_dir:  # Manual dir not passed
      return None

    if not expected_url_info or not expected_url_info.filename:
      return None  # Filename unknown.

    manual_path = self._manual_dir / expected_url_info.filename
    if not manual_path.exists():  # File not manually downloaded
      return None

    return manual_path

  def _get_checksum_dl_result(
      self, resource: resource_lib.Resource, legacy_mode: bool = False
  ) -> downloader.DownloadResult | None:
    """Checks if the download has been cached and checksum is known."""
    expected_url_info = self._url_infos.get(resource.url)

    if not expected_url_info:
      return None

    checksum_path = self._get_dl_path(
        resource, expected_url_info.checksum, legacy_mode=legacy_mode
    )
    if not resource_lib.is_locally_cached(checksum_path):
      return None

    return downloader.DownloadResult(
        path=checksum_path, url_info=expected_url_info
    )

  def _get_url_dl_result(
      self, resource: resource_lib.Resource, legacy_mode: bool = False
  ) -> downloader.DownloadResult | None:
    """Checks if the download has been cached and checksum is unknown."""
    url_path = self._get_dl_path(resource, legacy_mode=legacy_mode)
    if not resource_lib.is_locally_cached(url_path):
      return None

    expected_url_info = self._url_infos.get(resource.url)
    url_info = downloader.read_url_info(url_path)

    if expected_url_info and expected_url_info != url_info:
      # If checksums are registered but do not match, trigger a new
      # download (e.g. previous file corrupted, checksums updated)
      return None
    elif self._is_checksum_registered(url=resource.url):
      # Checksums were registered: Rename -> checksum_path
      path = self._get_dl_path(resource, url_info.checksum)
      resource_lib.replace_info_file(url_path, path)
      url_path.replace(path)
      return downloader.DownloadResult(path=path, url_info=url_info)
    else:
      # Checksums not registered: -> do nothing
      return downloader.DownloadResult(path=url_path, url_info=url_info)

  # Synchronize and memoize decorators ensure same resource will only be
  # processed once, even if passed twice to download_manager.
  @utils.build_synchronize_decorator()
  @utils.memoize()
  def _download_or_get_cache(
      self, resource: Url
  ) -> promise.Promise[epath.Path]:
    """Downloads resource or gets downloaded cache.

    Args:
      resource: The URL to download.

    Returns:
      Promise of the path to the downloaded resource.
    """
    # Normalize the input
    if not isinstance(resource, resource_lib.Resource):
      resource = resource_lib.Resource(url=resource)
    url = resource.url

    expected_url_info = self._url_infos.get(url)
    if (
        not self._register_checksums
        and self._force_checksums_validation
        and not expected_url_info
    ):
      raise ValueError(
          f'Missing checksums url: {url}, yet'
          ' `force_checksums_validation=True`. Did you forget to register'
          ' checksums?'
      )

    # User has manually downloaded the file.
    if manually_downloaded_path := self._get_manually_downloaded_path(
        expected_url_info
    ):
      computed_url_info = checksums.compute_url_info(manually_downloaded_path)
      dl_result = downloader.DownloadResult(
          path=manually_downloaded_path, url_info=computed_url_info
      )

    # Force download
    elif self._force_download:
      dl_result = None

    # Download has been cached (checksum known)
    elif dl_result := self._get_checksum_dl_result(resource):
      pass

    # Download has been cached (checksum known, legacy mode)
    elif dl_result := self._get_checksum_dl_result(resource, legacy_mode=True):
      pass

    # Download has been cached (checksum unknown)
    elif dl_result := self._get_url_dl_result(resource):
      pass

    # Download has been cached (checksum unknown, legacy mode)
    elif dl_result := self._get_url_dl_result(resource, legacy_mode=True):
      pass

    # Cache not found
    else:
      dl_result = None

    if dl_result:
      path = dl_result.path
      url_info = dl_result.url_info

      self._log_skip_download(url=url, url_info=url_info, path=path)
      self._register_or_validate_checksums(
          url=url, url_info=url_info, path=path
      )
      return promise.Promise.resolve(path)
    else:
      return self._download(resource)

  def _log_skip_download(
      self, url: str, url_info: checksums.UrlInfo, path: epath.Path
  ) -> None:
    logging.info(f'Skipping download of {url}: File cached in {path}')
    # Still update the progression bar to indicate the file was downloaded
    self._downloader.increase_tqdm(url_info)

  def _register_or_validate_checksums(
      self, url: str, url_info: checksums.UrlInfo, path: epath.Path
  ) -> None:
    """Registers or validates checksums depending on `self._register_checksums`."""
    # Used both in `.downloaded_size` and `_record_url_infos()`
    self._recorded_url_infos[url] = url_info

    if self._register_checksums:
      # Note:
      # * We save even if `expected_url_info == url_info` as
      #   `expected_url_info` might have been loaded from another dataset.
      # * `register_checksums_path` was validated in `__init__` so this
      #   shouldn't fail.
      self._record_url_infos()
    elif expected_url_info := self._url_infos.get(url):
      # Eventually validate checksums
      if expected_url_info != url_info:
        msg = (
            f'Artifact {url}, downloaded to {path}, has wrong checksum:\n'
            f'* Expected: {expected_url_info}\n'
            f'* Got: {url_info}\n'
            'To debug, see: '
            'https://www.tensorflow.org/datasets/overview#fixing_nonmatchingchecksumerror'
        )
        raise NonMatchingChecksumError(msg)

  def _is_checksum_registered(self, url: str) -> bool:
    """Returns whether checksums are registered for the given url."""
    if url in self._url_infos:
      # Checksum is already registered
      return True
    elif self._register_checksums:
      # Checksum is being registered
      return True
    else:
      # Checksum is not registered
      return False

  def _download(
      self, resource: resource_lib.Resource
  ) -> promise.Promise[epath.Path]:
    """Downloads resource.

    Args:
      resource: The resource to download.

    Returns:
      Promise of the path to the downloaded url.
    """
    url_path = self._get_dl_path(resource)
    url = resource.url

    # Download in a tmp directory next to url_path (to avoid name collisions)
    # `download_tmp_dir` is cleaned-up in `callback`
    download_tmp_dir = (
        url_path.parent / f'{url_path.name}.tmp.{uuid.uuid4().hex}'
    )
    download_tmp_dir.mkdir(parents=True, exist_ok=True)
    logging.info(f'Downloading {url} into {download_tmp_dir}...')
    future = self._downloader.download(
        url, download_tmp_dir, verify=self._verify_ssl
    )

    def callback(dl_result: downloader.DownloadResult) -> epath.Path:
      """Post-process the download result."""
      dl_path = dl_result.path
      dl_url_info = dl_result.url_info

      self._register_or_validate_checksums(
          url=url, url_info=dl_url_info, path=dl_path
      )
      if self._is_checksum_registered(url=url):
        dst_path = self._get_dl_path(resource, dl_url_info.checksum)
      else:
        dst_path = url_path

      resource_lib.write_info_file(
          url=url,
          path=dst_path,
          dataset_name=self._dataset_name,
          original_fname=dl_path.name,
          url_info=dl_url_info,
      )
      dl_path.replace(dst_path)
      dl_path.parent.rmdir()  # Cleanup tmp dir (will fail if dir not empty)

      return dst_path

    return future.then(callback)

  @utils.build_synchronize_decorator()
  @utils.memoize()
  def _extract(self, resource: ExtractPath) -> promise.Promise[epath.Path]:
    """Extract a single archive, returns Promise->path to extraction result."""
    if not isinstance(resource, resource_lib.Resource):
      resource = resource_lib.Resource(path=resource)
    path = resource.path
    extract_method = resource.extract_method
    if extract_method == resource_lib.ExtractMethod.NO_EXTRACT:
      logging.info('Skipping extraction for %s (method=NO_EXTRACT).', path)
      return promise.Promise.resolve(path)
    method_name = resource_lib.ExtractMethod(extract_method).name
    extract_path = self._extract_dir / f'{method_name}.{path.name}'
    if not self._force_extraction and extract_path.exists():
      logging.info('Reusing extraction of %s at %s.', path, extract_path)
      return promise.Promise.resolve(extract_path)
    return self._extractor.extract(path, extract_method, extract_path)

  @utils.build_synchronize_decorator()
  @utils.memoize()
  def _download_extract(self, resource):
    """Download-extract `Resource` or url, returns Promise->path."""
    if isinstance(resource, str):
      resource = resource_lib.Resource(url=resource)

    def callback(path):
      resource.path = path
      return self._extract(resource)

    return self._download_or_get_cache(resource).then(callback)

  def download_checksums(self, checksums_url):
    """Downloads checksum file from the given URL and adds it to registry."""
    checksums_path = self.download(checksums_url)
    self._url_infos.update(checksums.load_url_infos(checksums_path))

  def download_kaggle_data(self, competition_or_dataset: str) -> epath.Path:
    """Download data for a given Kaggle Dataset or competition.

    Note: This function requires the Kaggle CLI tool.
    Read the installation guide at https://www.kaggle.com/docs/api.

    Args:
      competition_or_dataset: Dataset name (`zillow/zecon`) or competition name
        (`titanic`)

    Returns:
      The path to the downloaded files.
    """
    return kaggle.download_kaggle_data(
        competition_or_dataset, self._download_dir
    )

  @typing.overload
  def download(self, url_or_urls: Url) -> epath.Path:
    ...

  @typing.overload
  def download(self, url_or_urls: dict[str, Url]) -> dict[str, epath.Path]:
    ...

  @typing.overload
  def download(self, url_or_urls: Tree[Url]) -> Tree[epath.Path]:
    ...

  def download(self, url_or_urls):
    """Download given url(s).

    Args:
      url_or_urls: url or `list`/`dict` of urls to download and extract. Each
        url can be a `str` or `tfds.download.Resource`.

    Returns:
      downloaded_path(s): `str`, The downloaded paths matching the given input
        url_or_urls.
    """
    # Add progress bar to follow the download state
    with self._downloader.tqdm():
      return _map_promise(self._download_or_get_cache, url_or_urls)

  def iter_archive(
      self,
      resource: ExtractPath,
  ) -> Iterator[tuple[str, typing.BinaryIO]]:
    """Returns iterator over files within archive.

    **Important Note**: caller should read files as they are yielded.
    Reading out of order is slow.

    Args:
      resource: path to archive or `tfds.download.Resource`.

    Returns:
      Generator yielding tuple (path_within_archive, file_obj).
    """
    if not isinstance(resource, resource_lib.Resource):
      resource = resource_lib.Resource(path=resource)
    return extractor.iter_archive(resource.path, resource.extract_method)

  @typing.overload
  def extract(self, path_or_paths: ExtractPath) -> epath.Path:
    ...

  @typing.overload
  def extract(
      self, path_or_paths: dict[str, ExtractPath]
  ) -> dict[str, epath.Path]:
    ...

  @typing.overload
  def extract(self, path_or_paths: Tree[ExtractPath]) -> Tree[epath.Path]:
    ...

  def extract(self, path_or_paths):
    """Extract given path(s).

    Args:
      path_or_paths: path or `list`/`dict` of path of file to extract. Each path
        can be a `str` or `tfds.download.Resource`.  If not explicitly specified
        in `Resource`, the extraction method is deduced from downloaded file
        name.

    Returns:
      extracted_path(s): `str`, The extracted paths matching the given input
        path_or_paths.
    """
    # Add progress bar to follow the download state
    with self._extractor.tqdm():
      return _map_promise(self._extract, path_or_paths)

  @typing.overload
  def download_and_extract(self, url_or_urls: Url) -> epath.Path:
    ...

  @typing.overload
  def download_and_extract(
      self, url_or_urls: dict[str, Url]
  ) -> dict[str, epath.Path]:
    ...

  @typing.overload
  def download_and_extract(self, url_or_urls: Tree[Url]) -> Tree[epath.Path]:
    ...

  def download_and_extract(self, url_or_urls):
    """Download and extract given url_or_urls.

    Is roughly equivalent to:

    ```
    extracted_paths = dl_manager.extract(dl_manager.download(url_or_urls))
    ```

    Args:
      url_or_urls: url or `list`/`dict` of urls to download and extract. Each
        url can be a `str` or `tfds.download.Resource`.  If not explicitly
        specified in `Resource`, the extraction method will automatically be
        deduced from downloaded file name.

    Returns:
      extracted_path(s): `str`, extracted paths of given URL(s).
    """
    # Add progress bar to follow the download state
    with self._downloader.tqdm():
      with self._extractor.tqdm():
        return _map_promise(self._download_extract, url_or_urls)

  @property
  def download_dir(self) -> epath.Path:
    return self._download_dir

  @functools.cached_property
  def manual_dir(self) -> epath.Path:
    """Returns the directory containing the manually extracted data."""
    if not self._manual_dir:
      raise AssertionError('Manual directory not enabled.')
    if not self._manual_dir_instructions:
      raise ValueError(
          'To access `dl_manager.manual_dir`, please set '
          '`MANUAL_DOWNLOAD_INSTRUCTIONS` in your dataset.'
      )
    if not self._manual_dir.exists() or not list(self._manual_dir.iterdir()):
      raise AssertionError(
          f'Manual directory {self._manual_dir} does not exist or is empty. '
          'Create it and download/extract dataset artifacts in there using '
          f'instructions:\n{self._manual_dir_instructions}'
      )
    return self._manual_dir


def _map_promise(map_fn, all_inputs):
  """Map the function into each element and resolve the promise."""
  all_promises = tree.map_structure(map_fn, all_inputs)  # Apply the function
  res = tree.map_structure(lambda p: p.get(), all_promises)  # Wait promises
  return res



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DatasetBuilder initialized from a Croissant config file.

Croissant is a high-level format for machine learning datasets
https://github.com/mlcommons/croissant

A CroissantBuilder initializes a TFDS DatasetBuilder from a Croissant dataset
file; each of the record_set_ids specified will result in a separate
ConfigBuilder.

```python
import tensorflow_datasets as tfds
builder = tfds.dataset_builders.CroissantBuilder(
    jsonld="https://raw.githubusercontent.com/mlcommons/croissant/main/datasets/1.0/huggingface-mnist/metadata.json",
    file_format='array_record',
)
builder.download_and_prepare()
ds = builder.as_data_source()
print(ds['default'][0])
```
"""

from __future__ import annotations

from collections.abc import Mapping, Sequence
import datetime
import json
from typing import Any

from etils import enp
from etils import epath
import numpy as np
from tensorflow_datasets.core import dataset_builder
from tensorflow_datasets.core import dataset_info
from tensorflow_datasets.core import download
from tensorflow_datasets.core import split_builder as split_builder_lib
from tensorflow_datasets.core import splits as splits_lib
from tensorflow_datasets.core.features import audio_feature
from tensorflow_datasets.core.features import bounding_boxes
from tensorflow_datasets.core.features import bounding_boxes_utils as bb_utils
from tensorflow_datasets.core.features import feature as feature_lib
from tensorflow_datasets.core.features import features_dict
from tensorflow_datasets.core.features import image_feature
from tensorflow_datasets.core.features import sequence_feature
from tensorflow_datasets.core.features import tensor_feature
from tensorflow_datasets.core.features import text_feature
from tensorflow_datasets.core.features import video_feature
from tensorflow_datasets.core.utils import conversion_utils
from tensorflow_datasets.core.utils import croissant_utils
from tensorflow_datasets.core.utils import type_utils
from tensorflow_datasets.core.utils import version as version_lib
from tensorflow_datasets.core.utils.lazy_imports_utils import apache_beam as beam
from tensorflow_datasets.core.utils.lazy_imports_utils import mlcroissant as mlc
from tensorflow_datasets.core.utils.lazy_imports_utils import pandas as pd


_RecordOrFeature = Mapping[str, Any]


def _strip_record_set_prefix(
    record_or_feature: _RecordOrFeature, record_set_id: str
) -> _RecordOrFeature:
  """Removes the record set prefix from the field ids of a record or feature."""
  return {
      field_id.removeprefix(f'{record_set_id}/'): value
      for field_id, value in record_or_feature.items()
  }


def array_datatype_converter(
    feature: type_utils.TfdsDType | feature_lib.FeatureConnector | None,
    field: mlc.Field,
    dtype_mapping: Mapping[type_utils.TfdsDType, type_utils.TfdsDType],
    language: str | None = None,
):
  """Includes the given feature in a sequence or tensor feature.

  Single-dimensional arrays are converted to sequences. Multi-dimensional arrays
  with unknown dimensions, or with non-native dtypes are converted to sequences
  of sequences. Otherwise, they are converted to tensors.

  Args:
    feature: The inner feature to include in a sequence or tensor feature.
    field: The mlc.Field object.
    dtype_mapping: A mapping of dtypes to the corresponding dtypes that will be
      used in TFDS.
    language: For Croissant jsonld which include multi-lingual descriptions, the
      language code to use to extract the description to be used in TFDS. If
      None, it will extract the description in English or the first available
      language in the dictionary.

  Returns:
    A sequence or tensor feature including the inner feature.
  """
  field_dtype = None
  if field.data_type in dtype_mapping:
    field_dtype = dtype_mapping[field.data_type]
  elif enp.lazy.is_np_dtype(field.data_type):
    field_dtype = field.data_type

  description = croissant_utils.extract_localized_string(
      field.description, language=language, field_name='description'
  )

  if len(field.array_shape_tuple) == 1:
    return sequence_feature.Sequence(feature, doc=description)
  elif (-1 in field.array_shape_tuple) or (field_dtype is None):
    for _ in range(len(field.array_shape_tuple)):
      feature = sequence_feature.Sequence(feature, doc=description)
    return feature
  else:
    return tensor_feature.Tensor(
        shape=field.array_shape_tuple,
        dtype=field_dtype,
        doc=description,
    )


def datatype_converter(
    field: mlc.Field,
    int_dtype: type_utils.TfdsDType = np.int64,
    float_dtype: type_utils.TfdsDType = np.float32,
    language: str | None = None,
):
  """Converts a Croissant field to a TFDS-compatible feature.

  Args:
    field: A mlcroissant Field object.
    int_dtype: The dtype to use for TFDS integer features. Defaults to np.int64.
    float_dtype: The dtype to use for TFDS float features. Defaults to
      np.float32.
    language: For Croissant jsonld which include multi-lingual descriptions, the
      language code to use to extract the description to be used in TFDS. If
      None, it will extract the description in English or the first available
      language in the dictionary.

  Returns:
    Converted datatype for TFDS, or None when a Field does not specify a type.

  Raises:
    NotImplementedError when the feature is not supported yet, or ValueError
    when a Field is malformed.
  """
  if field.is_enumeration:
    raise NotImplementedError('Not implemented yet.')
  dtype_mapping = {
      bool: np.bool_,
      bytes: np.str_,
      float: float_dtype,
      int: int_dtype,
  }

  field_data_type = field.data_type
  description = croissant_utils.extract_localized_string(
      field.description, language=language, field_name='description'
  )

  if not field_data_type:
    # Fields with sub fields are of type None.
    if field.sub_fields:
      feature = features_dict.FeaturesDict(
          {
              subfield.id: datatype_converter(
                  subfield,
                  int_dtype=int_dtype,
                  float_dtype=float_dtype,
                  language=language,
              )
              for subfield in field.sub_fields
          },
          doc=description,
      )
    else:
      feature = None
  elif field_data_type == bytes:
    feature = text_feature.Text(doc=description)
  elif field_data_type in dtype_mapping:
    feature = dtype_mapping[field_data_type]
  elif enp.lazy.is_np_dtype(field_data_type):
    feature = field_data_type
  # We return a text feature for date-time features (mlc.DataType.DATE,
  # mlc.DataType.DATETIME, and mlc.DataType.TIME).
  elif field_data_type == pd.Timestamp or field_data_type == datetime.time:
    feature = text_feature.Text(doc=description)
  elif field_data_type == mlc.DataType.IMAGE_OBJECT:
    feature = image_feature.Image(doc=description)
  elif field_data_type == mlc.DataType.BOUNDING_BOX:
    # TFDS uses REL_YXYX by default, but Hugging Face doesn't enforce a format.
    if bbox_format := field.source.format:
      try:
        bbox_format = bb_utils.BBoxFormat(bbox_format)
      except ValueError as e:
        raise ValueError(
            f'Unsupported bounding box format: {bbox_format}. Currently'
            ' supported bounding box formats are: '
            f'{[format.value for format in bb_utils.BBoxFormat]}'
        ) from e
    feature = bounding_boxes.BBoxFeature(
        doc=description, bbox_format=bbox_format
    )
  elif field_data_type == mlc.DataType.AUDIO_OBJECT:
    feature = audio_feature.Audio(
        doc=description, sample_rate=field.source.sampling_rate
    )
  elif field_data_type == mlc.DataType.VIDEO_OBJECT:
    feature = video_feature.Video(doc=description)
  else:
    raise ValueError(
        f'Unknown data type: {field_data_type} for field {field.id}.'
    )

  if feature and field.is_array:
    feature = array_datatype_converter(
        feature=feature,
        field=field,
        dtype_mapping=dtype_mapping,
        language=language,
    )
  # If the field is repeated, we return a sequence feature. `field.repeated` is
  # deprecated starting from Croissant 1.1, but we still support it for
  # backwards compatibility.
  if feature and field.repeated:
    feature = sequence_feature.Sequence(feature, doc=field.description)
  return feature


def _extract_license(license_: Any) -> str:
  """Extracts the full terms of a license as a string.

  In case the license is a CreativeWork, we join the name, description and url
  fields with brackets, e.g.
  [U.S. Government Works][https://www.usa.gov/government-works/].

  Args:
    license_: The license from mlcroissant.

  Returns:
    The full terms of the license as a string.
  """
  if isinstance(license_, str):
    return license_
  elif isinstance(license_, mlc.CreativeWork):
    possible_fields = [license_.name, license_.description, license_.url]
    fields = [field for field in possible_fields if field]
    return '[' + ']['.join(fields) + ']'
  raise ValueError(
      'license_ should be mlc.CreativeWork | str. Got'
      f' {type(license_)}: {license_}.'
  )


def _get_license(metadata: Any) -> str | None:
  """Gets the license from the metadata (if any) else returns None."""
  if not isinstance(metadata, mlc.Metadata):
    raise ValueError(f'metadata should be mlc.Metadata. Got {type(metadata)}')
  licenses = metadata.license
  if licenses:
    return ', '.join([_extract_license(l) for l in licenses if l])
  return None


class CroissantBuilder(
    dataset_builder.GeneratorBasedBuilder, skip_registration=True
):
  """DatasetBuilder initialized from a Croissant config file."""

  def __init__(
      self,
      *,
      jsonld: epath.PathLike | Mapping[str, Any],
      record_set_ids: Sequence[str] | None = None,
      disable_shuffling: bool | None = False,
      int_dtype: type_utils.TfdsDType = np.int64,
      float_dtype: type_utils.TfdsDType = np.float32,
      mapping: Mapping[str, epath.PathLike] | None = None,
      overwrite_version: version_lib.VersionOrStr | None = None,
      filters: Mapping[str, Any] | None = None,
      **kwargs: Any,
  ):
    """Initializes a CroissantBuilder.

    Args:
      jsonld: The Croissant JSON-LD for the given dataset: either a file path or
        a URL.
      record_set_ids: The @ids of the record sets for the dataset. Each record
        set will correspond to a separate config. If not specified, a config
        will be generated for each record set defined in the Croissant JSON-LD,
        except for the record sets which specify `cr:data`.
      disable_shuffling: Specify whether to shuffle the examples.
      int_dtype: The dtype to use for TFDS integer features. Defaults to
        np.int64.
      float_dtype: The dtype to use for TFDS float features. Defaults to
        np.float32.
      mapping: Mapping filename->filepath as a Python dict[str, str] to handle
        manual downloads. If `document.csv` is the FileObject and you downloaded
        it to `~/Downloads/document.csv`, you can specify
        `mapping={"document.csv": "~/Downloads/document.csv"}`.
      overwrite_version: Semantic version of the dataset to be set.
      filters: A dict of filters to apply to the records at preparation time (in
        the `_generate_examples` function). The keys should be field names and
        the values should be the values to filter by. If a record matches all
        the filters, it will be included in the dataset.
      **kwargs: kwargs to pass to GeneratorBasedBuilder directly.

    Raises:
      ValueError: If no record sets are found in the Croissant JSON-LD.
    """
    if mapping is None:
      mapping = {}
    self.jsonld = jsonld
    self.mapping = mapping
    dataset = mlc.Dataset(jsonld, mapping=mapping)
    self.name = croissant_utils.get_tfds_dataset_name(dataset)
    self.metadata = dataset.metadata

    # The dataset version is determined using the following precedence:
    # * overwrite_version (if provided).
    # * The version from Croissant metadata (self.metadata.version),
    # automatically converting major.minor formats to major.minor.0 (e.g., "1.2"
    # becomes "1.2.0"). See croissant_utils.get_croissant_version for details.
    # * Defaults to '1.0.0' if no version is specified (version is optional in
    # Croissant, but mandatory in TFDS).
    self.VERSION = version_lib.Version(  # pylint: disable=invalid-name
        overwrite_version
        or croissant_utils.get_croissant_version(self.metadata.version)
        or '1.0.0'
    )
    self.RELEASE_NOTES = {}  # pylint: disable=invalid-name

    if not record_set_ids:
      record_set_ids = croissant_utils.get_record_set_ids(self.metadata)
    config_names = [
        conversion_utils.to_tfds_name(record_set_id)
        for record_set_id in record_set_ids
    ]
    if not config_names:
      raise ValueError(
          'No record sets found in the Croissant JSON-LD. At least one record'
          ' set is required to be able to download and prepare the dataset.'
      )

    self.BUILDER_CONFIGS: list[dataset_builder.BuilderConfig] = [  # pylint: disable=invalid-name
        dataset_builder.BuilderConfig(name=config_name)
        for config_name in config_names
    ]

    self._disable_shuffling = disable_shuffling

    self._int_dtype = int_dtype
    self._float_dtype = float_dtype
    self._filters = filters or {}

    super().__init__(
        **kwargs,
    )

  @property
  def builder_config(self) -> dataset_builder.BuilderConfig:
    """`tfds.core.BuilderConfig` for this builder."""
    return (
        self._builder_config
    )  # pytype: disable=bad-return-type  # always-use-return-annotations

  def _info(self) -> dataset_info.DatasetInfo:
    return dataset_info.DatasetInfo(
        builder=self,
        description=self.metadata.description,
        features=self.get_features(),
        homepage=self.metadata.url,
        citation=self.metadata.cite_as,
        license=_get_license(self.metadata),
        disable_shuffling=self._disable_shuffling,
    )

  def get_features(self) -> features_dict.FeaturesDict:
    """Infers the features dict for the required record set."""
    record_set = croissant_utils.get_record_set(
        self.builder_config.name, metadata=self.metadata
    )
    fields = record_set.fields
    features = {}
    for field in fields:
      feature = datatype_converter(
          field, int_dtype=self._int_dtype, float_dtype=self._float_dtype
      )
      features[field.id] = feature
    features = _strip_record_set_prefix(features, record_set.id)
    return features_dict.FeaturesDict(features)

  def _split_generators(
      self,
      dl_manager: download.DownloadManager,
      pipeline: beam.Pipeline,
  ) -> dict[splits_lib.Split, split_builder_lib.SplitGenerator]:
    del dl_manager  # unused
    del pipeline  # unused
    # If a split recordset is joined for the required record set, we generate
    # splits accordingly. Otherwise, it generates a single `default` split with
    # all the records.
    record_set = croissant_utils.get_record_set(
        self.builder_config.name, metadata=self.metadata
    )
    if split_reference := croissant_utils.get_split_recordset(
        record_set, metadata=self.metadata
    ):
      # The key used in the split recordset's data is referenced in the
      # reference field.
      split_key = split_reference.reference_field.references.field
      return {
          split[split_key]: self._generate_examples(
              filters={
                  **self._filters,
                  split_reference.reference_field.id: split[split_key],
              },
          )
          for split in split_reference.split_record_set.data
      }
    else:
      return {'default': self._generate_examples(filters=self._filters)}

  def _generate_examples(
      self,
      filters: dict[str, Any],
  ) -> beam.PTransform:
    """Generates the examples for the given record set.

    Args:
      filters: A dict of filters to apply to the records. The keys should be
        field names and the values should be the values to filter by. If a
        record matches all the filters, it will be included in the dataset.

    Returns:
      A collection with tuple of (index, record) for each record in the dataset.
    """
    record_set = croissant_utils.get_record_set(
        self.builder_config.name, metadata=self.metadata
    )
    dataset = mlc.Dataset(self.jsonld, mapping=self.mapping)
    records = dataset.records(record_set.id, filters=filters)

    def convert_to_tfds_format(
        global_index: int,
        record: _RecordOrFeature,
        features: feature_lib.FeatureConnector | None = None,
        record_set_id: str | None = None,
    ) -> tuple[int, _RecordOrFeature]:
      if not features:
        raise ValueError('features should not be None.')
      if not record_set_id:
        raise ValueError('record_set_id should not be None.')
      record = _strip_record_set_prefix(record, record_set_id)
      return (
          global_index,
          conversion_utils.to_tfds_value(record, features),
      )

    return (
        records.beam_reader()
        | f'Convert to TFDS format for filters: {json.dumps(filters)}'
        >> beam.MapTuple(
            convert_to_tfds_format,
            features=self.info.features,
            record_set_id=record_set.id,
        )
    )



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for bounding_boxes_utils."""

import numpy as np
import pytest
from tensorflow_datasets.core.features import bounding_boxes_utils as bb_utils


TEST_INPUT_LIST = [199.84, 200.46, 270.72, 278.17]


@pytest.mark.parametrize(
    'test_input,input_format,img_shape,normalize,expected',
    [
        (TEST_INPUT_LIST, 'YXYX', None, False, TEST_INPUT_LIST),
        (
            TEST_INPUT_LIST,
            bb_utils.BBoxFormat.YXYX,
            None,
            False,
            TEST_INPUT_LIST,
        ),
        (
            TEST_INPUT_LIST,
            'XYXY',
            None,
            False,
            [200.46, 199.84, 278.17, 270.72],
        ),
        ([200.46, 199.84, 77.71, 70.88], 'XYWH', None, False, TEST_INPUT_LIST),
    ],
)
def test_convert_to_bbox(
    test_input, input_format, img_shape, normalize, expected
):
  actual = bb_utils.convert_coordinates_to_bbox(
      coordinates=np.array(test_input),
      input_format=input_format,
      img_shape=img_shape,
      normalize=normalize,
  )
  expected_bbox = bb_utils.BBox(
      ymin=expected[0],
      xmin=expected[1],
      ymax=expected[2],
      xmax=expected[3],
  )
  assert actual == expected_bbox


@pytest.mark.parametrize(
    'test_input,input_format,error_match',
    [
        ([1, 2, 3], None, 'Expected 4'),
        (TEST_INPUT_LIST, 'NonExistentConverter', 'Unsupported bbox format'),
        (
            TEST_INPUT_LIST,
            bb_utils.BBoxFormat.XYXY,
            'If normalize is True, img_shape must be provided, but got None.',
        ),
        (
            TEST_INPUT_LIST,
            bb_utils.BBoxFormat.REL_XYXY,
            (
                'If the input format is normalized, then normalize should be'
                ' False.'
            ),
        ),
    ],
)
def test_convert_coordinates_to_bbox_valueerror(
    test_input, input_format, error_match
):
  with pytest.raises(ValueError, match=rf'{error_match}'):
    bb_utils.convert_coordinates_to_bbox(
        coordinates=np.array(test_input), input_format=input_format
    )


@pytest.mark.parametrize(
    'bbox,img_shape,expected',
    [
        ([10, 20, 110, 120], [100, 200], [0.1, 0.1, 1.1, 0.6]),
        ([10, 20, 110, 120], np.array([1000, 2000]), [0.01, 0.01, 0.11, 0.06]),
    ],
)
def test_normalize_bbox(bbox, img_shape, expected):
  bbox = bb_utils.BBox(ymin=bbox[0], xmin=bbox[1], ymax=bbox[2], xmax=bbox[3])
  actual = bb_utils.normalize_bbox(bbox, img_shape)
  expected_bbox = bb_utils.BBox(
      ymin=expected[0], xmin=expected[1], ymax=expected[2], xmax=expected[3]
  )
  assert actual == expected_bbox


@pytest.mark.parametrize(
    'bbox,img_shape,match',
    [
        (TEST_INPUT_LIST, [1], 'Expected 2'),
        (TEST_INPUT_LIST, np.array([1, 2, 3]), 'Expected 2'),
        (TEST_INPUT_LIST, [0, 0], 'shapes cannot be zero'),
    ],
)
def test_normalize_bbox_valueerror(bbox, img_shape, match):
  with pytest.raises(ValueError, match=rf'{match}'):
    bb_utils.normalize_bbox(bbox, img_shape)



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""smartwatch_gestures dataset."""

from __future__ import annotations

import numpy as np
import tensorflow_datasets.public_api as tfds


class Builder(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for smartwatch_gestures dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    class_label = tfds.features.ClassLabel(
        names_file=tfds.core.tfds_path(
            'datasets/smartwatch_gestures/labels.txt'
        )
    )
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            'features': tfds.features.Sequence({
                'time_millis': np.uint64,
                'time_nanos': np.uint64,
                'time_event': np.uint64,
                'accel_x': np.float64,
                'accel_y': np.float64,
                'accel_z': np.float64,
            }),
            'participant': np.uint8,
            'attempt': np.uint8,
            'gesture': class_label,
        }),
        supervised_keys=('features', 'gesture'),
        homepage='https://tev.fbk.eu/resources/smartwatch',
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""
    path = dl_manager.download_and_extract(
        'https://drive.google.com/uc?export=download&'
        'id=1nEs-JlAQv6xpuSIqahTKK68TgK37GirP'
    )

    # There are no predefined train/val/test split for this dataset.
    # (smartwatch_gestures): Returns the Dict['train', Iterator[Key, Example]].
    return {
        'train': self._generate_examples(path / 'gestures-dataset'),
    }

  def _generate_examples(self, path):
    """Yields examples."""

    pd = tfds.core.lazy_imports.pandas

    for f in path.glob('*/*/*.txt'):
      table = pd.read_table(
          f,
          sep=' ',
          header=None,
          names=[
              'time_millis',
              'time_nanos',
              'time_event',
              'accel_x',
              'accel_y',
              'accel_z',
          ],
      )

      # Create a unique key for each recorded gesture.
      participant_numstr = f.parent.parent.name[1:]  # Drop the "U".
      gesture_numstr = f.parent.name
      attempt_numstr = f.stem
      k = int(''.join([participant_numstr, gesture_numstr, attempt_numstr]))

      yield k, {
          'features': {
              'time_millis': table['time_millis'].to_numpy(),
              'time_nanos': table['time_nanos'].to_numpy(),
              'time_event': table['time_event'].to_numpy(),
              'accel_x': table['accel_x'].to_numpy(),
              'accel_y': table['accel_y'].to_numpy(),
              'accel_z': table['accel_z'].to_numpy(),
          },
          'participant': int(participant_numstr),
          'attempt': int(attempt_numstr),
          'gesture': int(gesture_numstr) - 1,
      }



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""robomimic_mg dataset."""

from __future__ import annotations

import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.robomimic import dataset_utils as utils


class Builder(utils.RobomimicBuilder):
  """DatasetBuilder for robomimic_mg dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }
  BUILDER_CONFIGS = utils.make_builder_configs(utils.DataSource.MG)
  DATASET_NAME = 'robomimic_mg'
  DATASET_FILE_EXTENSION: str = '_sparse'



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import textwrap

from etils import epath
import pytest
from tensorflow_datasets.core.dataset_builders.conll import conll_dataset_builder
from tensorflow_datasets.core.dataset_builders.conll import conll_dataset_builder_utils as conll_lib
import tensorflow_datasets.public_api as tfds

_FOLDER_PATH = "mock/path"

_VALID_INPUT = textwrap.dedent("""
-DOCSTART- -X- -X- O
Winter NN B-NP O
is VBZ B-VP O

Air NN I-NP O
. . O O
""")

_INVALID_INPUT = textwrap.dedent("""
Winter NN B-NP
is VBZ B-VP O

Air NN I-NP O
. . O O
""")


class DummyConllDataset(conll_dataset_builder.ConllDatasetBuilder):
  VERSION = tfds.core.Version("1.0.0")
  RELEASE_NOTES = {"1.0.0": "Dummy notes."}
  BUILDER_CONFIGS = [conll_lib.CONLL_2003_CONFIG]

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return self.create_dataset_info(
        description="Dummy CoNLL dataset.",
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    del dl_manager
    return {"train": self._generate_examples("/tmp/input.txt")}


def test_generate_example(tmpdir):
  tmpdir = epath.Path(tmpdir)
  input_path = tmpdir / "input_path.txt"
  input_path.write_text(_VALID_INPUT)

  dataset = DummyConllDataset(data_dir=tmpdir)
  examples = list(dataset._generate_examples(input_path))

  expected_examples = [
      (
          0,
          {
              "tokens": ["Winter", "is"],
              "pos": ["NN", "VBZ"],
              "chunks": ["B-NP", "B-VP"],
              "ner": ["O", "O"],
          },
      ),
      (
          1,
          {
              "tokens": ["Air", "."],
              "pos": ["NN", "."],
              "chunks": ["I-NP", "O"],
              "ner": ["O", "O"],
          },
      ),
  ]

  assert examples == expected_examples

  for _, example in examples:
    assert len(example) == len(conll_lib.CONLL_2003_ORDERED_FEATURES)

  assert len(examples) == 2


def test_generate_corrupted_example(tmpdir):
  tmpdir = epath.Path(tmpdir)
  input_path = tmpdir / "input_path.txt"
  input_path.write_text(_INVALID_INPUT)
  dataset = DummyConllDataset(data_dir=tmpdir)

  error_line = "Winter NN B-NP"
  error_msg = (
      f"Mismatch in the number of features found in line: {error_line}\n\n"
      "Should be 4, but found 3"
  )
  with pytest.raises(ValueError, match=error_msg):
    list(dataset._generate_examples(input_path))



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""robomimic_mh dataset."""

from __future__ import annotations

from typing import Any, Dict

import numpy as np
import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.robomimic import dataset_utils as utils


class Builder(utils.RobomimicBuilder):
  """DatasetBuilder for robomimic_mh dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }
  BUILDER_CONFIGS = utils.make_builder_configs(utils.DataSource.MH)
  DATASET_NAME = 'robomimic_mh'

  def _get_metadata(self) -> Dict[Any, Any]:
    if self.builder_config.task == utils.Task.TRANSPORT:
      episode_metadata = {
          '20_percent': np.bool_,
          '20_percent_train': np.bool_,
          '20_percent_valid': np.bool_,
          '50_percent': np.bool_,
          '50_percent_train': np.bool_,
          '50_percent_valid': np.bool_,
          'better': np.bool_,
          'better_train': np.bool_,
          'better_valid': np.bool_,
          'okay': np.bool_,
          'okay_better': np.bool_,
          'okay_better_train': np.bool_,
          'okay_better_valid': np.bool_,
          'okay_train': np.bool_,
          'okay_valid': np.bool_,
          'train': np.bool_,
          'valid': np.bool_,
          'worse': np.bool_,
          'worse_better': np.bool_,
          'worse_better_train': np.bool_,
          'worse_better_valid': np.bool_,
          'worse_okay': np.bool_,
          'worse_okay_train': np.bool_,
          'worse_okay_valid': np.bool_,
          'worse_train': np.bool_,
          'worse_valid': np.bool_,
      }
    else:
      episode_metadata = {
          '20_percent': np.bool_,
          '20_percent_train': np.bool_,
          '20_percent_valid': np.bool_,
          '50_percent': np.bool_,
          '50_percent_train': np.bool_,
          '50_percent_valid': np.bool_,
          'better': np.bool_,
          'better_operator_1': np.bool_,
          'better_operator_1_train': np.bool_,
          'better_operator_1_valid': np.bool_,
          'better_operator_2': np.bool_,
          'better_operator_2_train': np.bool_,
          'better_operator_2_valid': np.bool_,
          'better_train': np.bool_,
          'better_valid': np.bool_,
          'okay': np.bool_,
          'okay_better': np.bool_,
          'okay_better_train': np.bool_,
          'okay_better_valid': np.bool_,
          'okay_operator_1': np.bool_,
          'okay_operator_1_train': np.bool_,
          'okay_operator_1_valid': np.bool_,
          'okay_operator_2': np.bool_,
          'okay_operator_2_train': np.bool_,
          'okay_operator_2_valid': np.bool_,
          'okay_train': np.bool_,
          'okay_valid': np.bool_,
          'train': np.bool_,
          'valid': np.bool_,
          'worse': np.bool_,
          'worse_better': np.bool_,
          'worse_better_train': np.bool_,
          'worse_better_valid': np.bool_,
          'worse_okay': np.bool_,
          'worse_okay_train': np.bool_,
          'worse_okay_valid': np.bool_,
          'worse_operator_1': np.bool_,
          'worse_operator_1_train': np.bool_,
          'worse_operator_1_valid': np.bool_,
          'worse_operator_2': np.bool_,
          'worse_operator_2_train': np.bool_,
          'worse_operator_2_valid': np.bool_,
          'worse_train': np.bool_,
          'worse_valid': np.bool_,
      }
    return episode_metadata



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""webvid dataset."""

import numpy as np

from tensorflow_datasets.datasets.webvid import webvid_dataset_builder
import tensorflow_datasets.public_api as tfds


def fake_ffmpeg_decode_diskless(*_, **__):
  img_size = webvid_dataset_builder._IMG_SIZE
  return np.ones((11, img_size[0], img_size[1], 3), dtype=np.uint8)


class WebvidTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for webvid dataset."""

  DATASET_CLASS = webvid_dataset_builder.Builder
  SPLITS = {
      'train': 3,  # Number of fake train examples.
      'validation': 1,  # Number of fake test examples.
  }

  def setUp(self):
    webvid_dataset_builder._ffmpeg_decode_diskless = fake_ffmpeg_decode_diskless
    super().setUp()


if __name__ == '__main__':
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""ogbg_molpcba dataset."""

from typing import Dict, Text, Tuple

from etils import epath
import numpy as np
import tensorflow_datasets.public_api as tfds

# Type hints.
ArrayDict = Dict[Text, np.ndarray]
Path = epath.Path

# URL.
_OGB_URL = 'https://ogb.stanford.edu/docs/graphprop'
_DOWNLOAD_URL = (
    'https://snap.stanford.edu/ogb/data/graphproppred/csv_mol_download/pcba.zip'
)

# File containing the names of individual tasks.
_TASKS_FNAME = 'datasets/ogbg_molpcba/tasks.txt'


class Builder(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for ogbg_molpcba dataset."""

  VERSION = tfds.core.Version('0.1.3')
  RELEASE_NOTES = {
      '0.1.0': 'Initial release of experimental API.',
      '0.1.1': 'Exposes the number of edges in each graph explicitly.',
      '0.1.2': 'Add metadata field for GraphVisualizer.',
      '0.1.3': 'Add metadata field for names of individual tasks.',
  }

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    # Read the individual task names.
    tasks_file = tfds.core.tfds_path(_TASKS_FNAME)
    tasks = tasks_file.read_text().splitlines()

    # Specify the tfds.core.DatasetInfo object
    return self.dataset_info_from_configs(
        # We mimic the features of the OGB platform-agnostic DataLoader.
        features=tfds.features.FeaturesDict({
            'num_nodes': tfds.features.Tensor(shape=(None,), dtype=np.int64),
            'node_feat': tfds.features.Tensor(
                shape=(None, 9), dtype=np.float32
            ),
            'num_edges': tfds.features.Tensor(shape=(None,), dtype=np.int64),
            'edge_feat': tfds.features.Tensor(
                shape=(None, 3), dtype=np.float32
            ),
            'edge_index': tfds.features.Tensor(shape=(None, 2), dtype=np.int64),
            'labels': tfds.features.Tensor(shape=(128,), dtype=np.float32),
        }),
        supervised_keys=None,
        homepage=_OGB_URL,
        metadata=tfds.core.MetadataDict({
            'tasks': tasks,
            'graph_visualizer': tfds.visualization.GraphVisualizerMetadataDict(
                edgelist_feature_name='edge_index'
            ),
        }),
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    # Download the original data.
    path = dl_manager.download_and_extract(_DOWNLOAD_URL)

    # Read the extracted data.
    data_path = path / 'pcba/raw'
    split_path = path / 'pcba/split/scaffold'
    all_data, split_indices = _read_extracted_data(data_path, split_path)

    # Return a list of the train/validation/test split generators.
    return {
        tfds.Split.TRAIN: self._generate_examples(
            all_data, split_indices['train']
        ),
        tfds.Split.VALIDATION: self._generate_examples(
            all_data, split_indices['valid']
        ),
        tfds.Split.TEST: self._generate_examples(
            all_data, split_indices['test']
        ),
    }

  def _generate_examples(self, all_data: ArrayDict, split_indices: np.ndarray):
    """Yields examples."""
    # Precompute for later.
    num_total_graphs = len(all_data['labels'])
    split_indices = set(split_indices)
    accumulated_num_nodes = np.concatenate(
        [np.array([0]), np.cumsum(all_data['num_nodes'])]
    )
    accumulated_num_edges = np.concatenate(
        [np.array([0]), np.cumsum(all_data['num_edges'])]
    )

    # Loop over the training set.
    for idx in range(num_total_graphs):
      # Check if this example is part of the split.
      if idx not in split_indices:
        continue

      # Read all of the graph information.
      labels = all_data['labels'][idx]
      num_nodes = all_data['num_nodes'][idx]
      node_slice = slice(
          accumulated_num_nodes[idx], accumulated_num_nodes[idx + 1]
      )
      node_feat = all_data['node_feat'][node_slice]
      num_edges = all_data['num_edges'][idx]
      edge_slice = slice(
          accumulated_num_edges[idx], accumulated_num_edges[idx + 1]
      )
      edge_feat = all_data['edge_feat'][edge_slice]
      edge_index = all_data['edge_index'][edge_slice]

      # Combine into a single dictionary.
      record = {
          'labels': labels,
          'num_nodes': num_nodes,
          'node_feat': node_feat,
          'num_edges': num_edges,
          'edge_feat': edge_feat,
          'edge_index': edge_index,
      }
      yield idx, record


def _read_extracted_data(
    data_path: Path, split_path: Path
) -> Tuple[ArrayDict, ArrayDict]:
  """Reads and processes the extracted graph data and splits."""
  pd = tfds.core.lazy_imports.pandas

  # Load columns describing the graph features and structure.
  column_names = [
      'edge_index',
      'num_nodes',
      'num_edges',
      'node_feat',
      'edge_feat',
      'labels',
  ]
  file_names = [
      'edge.csv.gz',
      'num-node-list.csv.gz',
      'num-edge-list.csv.gz',
      'node-feat.csv.gz',
      'edge-feat.csv.gz',
      'graph-label.csv.gz',
  ]
  dtypes = [
      np.int64,
      np.int64,
      np.int64,
      np.float32,
      np.float32,
      np.float32,
  ]
  all_data = {}
  for column_name, file_name, dtype in zip(column_names, file_names, dtypes):
    with (data_path / file_name).open('rb') as fp:
      values = pd.read_csv(fp, compression='gzip', header=None).values
      values = values.astype(dtype)
      all_data[column_name] = values

  # Load data splits.
  split_indices = {}
  for split_name in ['train', 'valid', 'test']:
    with (split_path / ('%s.csv.gz' % split_name)).open('rb') as fp:
      indices = pd.read_csv(fp, compression='gzip', header=None).values.T[0]
      split_indices[split_name] = indices

  return all_data, split_indices



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Lost and Found Road Hazard Dataset."""

from os import path
import re

from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds


class LostAndFoundConfig(tfds.core.BuilderConfig):
  """BuilderConfig for 'Lost and Found'.

  Args:
    right_images (bool): Enables right images for stereo image tasks.
    segmentation_labels (bool): Enables image segmentation labels.
    instance_ids (bool): Enables instance-id labels.
    disparity_maps (bool): Enables disparity maps.
    use_16bit (bool): Loads 16 bit (rgb) images instead of 8bit.
  """

  def __init__(
      self,
      right_images=False,
      segmentation_labels=False,
      instance_ids=False,
      disparity_maps=False,
      use_16bit=False,
      **kwargs,
  ):
    super(LostAndFoundConfig, self).__init__(**kwargs)

    self.features = ['image_left']
    if right_images:
      self.features.append('image_right')
    if segmentation_labels:
      self.features.append('segmentation_label')
    if instance_ids:
      self.features.append('instance_id')
    if disparity_maps:
      self.features.append('disparity_map')

    self.left_image_string = 'leftImg{}bit'.format('16' if use_16bit else '8')
    self.right_image_string = 'rightImg{}bit'.format('16' if use_16bit else '8')


class Builder(tfds.core.GeneratorBasedBuilder):
  """Lost and Found Road Hazard Dataset."""

  VERSION = tfds.core.Version('1.0.0')

  BUILDER_CONFIGS = [
      LostAndFoundConfig(
          name='semantic_segmentation',
          description='Lost and Found semantic segmentation dataset.',
          version='1.0.0',
          right_images=False,
          segmentation_labels=True,
          instance_ids=False,
          disparity_maps=False,
          use_16bit=False,
      ),
      LostAndFoundConfig(
          name='stereo_disparity',
          description='Lost and Found stereo images and disparity maps.',
          version='1.0.0',
          right_images=True,
          segmentation_labels=False,
          instance_ids=False,
          disparity_maps=True,
          use_16bit=False,
      ),
      LostAndFoundConfig(
          name='full',
          description='Full Lost and Found dataset.',
          version='1.0.0',
          right_images=True,
          segmentation_labels=True,
          instance_ids=True,
          disparity_maps=True,
          use_16bit=False,
      ),
      LostAndFoundConfig(
          name='full_16bit',
          description='Full Lost and Found dataset.',
          version='1.0.0',
          right_images=True,
          segmentation_labels=True,
          instance_ids=True,
          disparity_maps=True,
          use_16bit=True,
      ),
  ]

  def _info(self):
    possible_features = {
        'image_left': tfds.features.Image(
            shape=(1024, 2048, 3), encoding_format='png'
        ),
        'image_right': tfds.features.Image(
            shape=(1024, 2048, 3), encoding_format='png'
        ),
        'segmentation_label': tfds.features.Image(
            shape=(1024, 2048, 1), encoding_format='png', use_colormap=True
        ),
        'instance_id': tfds.features.Image(
            shape=(1024, 2048, 1), encoding_format='png', use_colormap=True
        ),
        'disparity_map': tfds.features.Image(
            shape=(1024, 2048, 1), encoding_format='png'
        ),
    }
    features = {
        feat: possible_features[feat] for feat in self.builder_config.features
    }
    features['image_id'] = tfds.features.Text()
    features = tfds.features.FeaturesDict(features)
    return self.dataset_info_from_configs(
        # tfds.features.FeatureConnectors
        features=features,
        # Homepage of the dataset for documentation
        homepage='http://www.6d-vision.com/lostandfounddataset',
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""
    base_url = (
        'http://www.dhbw-stuttgart.de/~sgehrig/lostAndFoundDataset/{}.zip'
    )

    # For each feature, this is the name of the zipfile and
    # root-directory in the archive
    zip_file_names = {
        'image_left': self.builder_config.left_image_string,
        'image_right': self.builder_config.right_image_string,
        'segmentation_label': 'gtCoarse',
        'instance_id': 'gtCoarse',
        'disparity_map': 'disparity',
    }

    download_urls = {
        feat: base_url.format(zip_file_names[feat])
        for feat in self.builder_config.features
    }
    # Split download and extract in two functions such that mock-data can
    # replace the result of the download function and is still used as input to
    # extract. Therefore, fake_data can be compressed zip archives.
    dl_paths = dl_manager.extract(dl_manager.download(download_urls))

    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            # These kwargs will be passed to _generate_examples
            gen_kwargs={
                feat: path.join(dl_paths[feat], zip_file_names[feat], 'train')
                for feat in self.builder_config.features
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST,
            # These kwargs will be passed to _generate_examples
            gen_kwargs={
                feat: path.join(dl_paths[feat], zip_file_names[feat], 'test')
                for feat in self.builder_config.features
            },
        ),
    ]

  def _generate_examples(self, **paths):
    """Yields examples."""
    # different file-suffixes dependent on the feature to load
    file_suffix = {
        'image_left': self.builder_config.left_image_string,
        'image_right': self.builder_config.right_image_string,
        'segmentation_label': 'gtCoarse_labelIds',
        'instance_id': 'gtCoarse_instanceIds',
        'disparity_map': 'disparity',
    }

    for scene_id in tf.io.gfile.listdir(paths['image_left']):
      paths_city_root = {
          feat: path.join(feat_dir, scene_id)
          for feat, feat_dir in paths.items()
      }

      left_city_root = paths_city_root['image_left']
      for left_img in tf.io.gfile.listdir(left_city_root):
        image_id = _get_id_from_left_image(left_img)

        features = {
            feat: path.join(
                paths_city_root[feat],
                '{}_{}.png'.format(image_id, file_suffix[feat]),
            )
            for feat in paths
        }
        features['image_id'] = image_id

        yield image_id, features


# Helper functions

LEFT_IMAGE_FILE_RE = re.compile(r'(.+)_leftImg(?:8|16)bit\.png')


def _get_id_from_left_image(left_image):
  """Returns the id of an image file.

  Used to associate an image file
  with its corresponding label.
  Example:
    'bonn_000001_000019_leftImg8bit' -> 'bonn_000001_000019'

  Args:
    left_image: left image file name.

  Returns:
    id of the image.
  """
  return LEFT_IMAGE_FILE_RE.match(left_image).group(1)



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""story_cloze dataset."""

from __future__ import annotations

import csv
import os

from etils import epath
import numpy as np
import tensorflow_datasets.public_api as tfds


class Builder(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for story_cloze dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }
  BUILDER_CONFIGS = [
      tfds.core.BuilderConfig(name='2016', description='2018 year'),
      tfds.core.BuilderConfig(name='2018', description='2018 year'),
  ]
  MANUAL_DOWNLOAD_INSTRUCTIONS = """\
  Visit https://www.cs.rochester.edu/nlp/rocstories/ and fill out the google
  form to obtain the datasets. You will receive an email with the link to
  download the datasets. For the 2016 data, the validation and test file needs
  to be renamed to cloze_test_val__spring2016.csv and
  cloze_test_test__spring2016.csv respectively. For 2018 version, the validation
  and test file needs to be renamed to cloze_test_val__winter2018.csv and
  to cloze_test_test__winter2018.csv respectively. Move both these files
  to the manual directory.
  """

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            'context': tfds.features.Text(),
            'endings': tfds.features.Sequence(tfds.features.Text()),
            'label': np.int32,
        }),
        supervised_keys=None,  # e.g. ('image', 'label')
        homepage='https://www.cs.rochester.edu/nlp/rocstories/',
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    if self.builder_config.name == '2016':
      file_postfix = 'spring2016'
    elif self.builder_config.name == '2018':
      file_postfix = 'winter2018'

    val_file = 'cloze_test_val__' + file_postfix + '.csv'
    test_file = 'cloze_test_test__' + file_postfix + '.csv'

    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.VALIDATION,
            gen_kwargs={
                'filepath': os.path.join(dl_manager.manual_dir, val_file)
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST,
            gen_kwargs={
                'filepath': os.path.join(dl_manager.manual_dir, test_file)
            },
        ),
    ]

  def _generate_examples(self, filepath):
    """Yields examples."""
    with epath.Path(filepath).open() as f:
      reader = csv.DictReader(f)
      for row in reader:
        context = ' '.join([
            row['InputSentence1'],
            row['InputSentence2'],
            row['InputSentence3'],
            row['InputSentence4'],
        ])
        endings = [
            row['RandomFifthSentenceQuiz1'],
            row['RandomFifthSentenceQuiz2'],
        ]

        yield row['InputStoryid'], {
            'context': context,
            'endings': endings,
            'label': row.get('AnswerRightEnding', -1),
        }



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

import datasets as hf_datasets
import numpy as np
import pytest
from tensorflow_datasets.core import lazy_imports_lib
from tensorflow_datasets.core.dataset_builders import huggingface_dataset_builder
from tensorflow_datasets.core.utils.lazy_imports_utils import huggingface_hub

PIL_Image = lazy_imports_lib.lazy_imports.PIL_Image


class DummyHuggingfaceBuilder(hf_datasets.GeneratorBasedBuilder):

  def _info(self):
    return hf_datasets.DatasetInfo(
        description='description',
        citation='citation',
        license='test-license',
        features=None,
        version='1.0.0',
    )

  def _split_generators(self, dl_manager):
    return [hf_datasets.SplitGenerator(name='train.clean')]

  def _generate_examples(self):
    for i in range(2):
      yield i, {
          'number': i,
          'text': f'{i}',
          'image': PIL_Image.new(mode='L', size=(4, 4)),
      }


@pytest.fixture(name='load_dataset_builder')
def mock_load_dataset_builder(tmp_path):
  hf_builder = DummyHuggingfaceBuilder(cache_dir=tmp_path / 'data')
  hf_builder.download_and_prepare = mock.Mock(
      wraps=hf_builder.download_and_prepare
  )
  with mock.patch.object(
      hf_datasets, 'load_dataset_builder', return_value=hf_builder
  ) as load_dataset_builder:
    yield load_dataset_builder


@pytest.fixture(name='login_to_hf')
def mock_login_to_hf():
  with mock.patch.object(
      huggingface_dataset_builder, 'login_to_hf'
  ) as login_to_hf:
    yield login_to_hf


@pytest.fixture(autouse=True)
def mock_hub_dataset_info():
  fake_dataset_info = huggingface_hub.hf_api.DatasetInfo(
      id='foo/bar',
      citation='citation from the hub',
      private=False,
      downloads=123,
      likes=456,
      tags=[],
  )
  with mock.patch.object(
      huggingface_hub, 'dataset_info', return_value=fake_dataset_info
  ) as dataset_info:
    yield dataset_info


@pytest.fixture(name='builder')
def mock_huggingface_dataset_builder(
    tmp_path, load_dataset_builder, login_to_hf
):
  builder = huggingface_dataset_builder.HuggingfaceDatasetBuilder(
      file_format='array_record',
      hf_repo_id='foo/bar',
      hf_config='config',
      ignore_verifications=True,
      data_dir=tmp_path / 'data',
      hf_hub_token='SECRET_TOKEN',
      hf_num_proc=100,
      other_arg='this is another arg',
  )
  load_dataset_builder.assert_called_once_with(
      'foo/bar', 'config', other_arg='this is another arg'
  )
  login_to_hf.assert_called_once_with('SECRET_TOKEN')
  yield builder


def test_dataset_info(builder):
  assert builder.info.description == 'description'
  assert builder.info.citation == 'citation from the hub'
  assert builder.info.redistribution_info.license == 'test-license'
  assert builder.info.homepage == 'https://huggingface.co/datasets/foo/bar'


def test_download_and_prepare(builder):
  builder.download_and_prepare()
  ds = builder.as_data_source()
  expected_image = PIL_Image.new(mode='RGB', size=(4, 4))
  # Split names are sanitized, eg train.clean -> train_clean
  for element, expected in zip(
      ds['train_clean'],
      [
          {'number': 0, 'text': b'0', 'image': expected_image},
          {'number': 1, 'text': b'1', 'image': expected_image},
      ],
  ):
    for feature in ['number', 'text', 'image']:
      assert np.array_equal(element[feature], expected[feature])
  assert len(ds['train_clean']) == 2


def test_all_parameters_are_passed_down_to_hf(builder):
  builder._hf_builder.download_and_prepare.assert_called_once_with(
      verification_mode='no_checks', num_proc=100
  )


def test_hf_features(builder):
  assert builder._hf_features() == {
      'number': hf_datasets.Value('int64'),
      'text': hf_datasets.Value('string'),
      'image': hf_datasets.Image(),
  }



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions for TFDS CLI."""

import abc
import argparse
from collections.abc import Callable, Sequence
import dataclasses
import itertools
import pathlib
from typing import TypeVar

from absl import logging
from absl.flags import argparse_flags
from etils import epath
import simple_parsing
from tensorflow_datasets.core import constants
from tensorflow_datasets.core import dataset_builder
from tensorflow_datasets.core import download
from tensorflow_datasets.core import file_adapters
from tensorflow_datasets.core import naming
from tensorflow_datasets.core.utils import file_utils
from tensorflow_datasets.scripts.utils import flag_utils

_DataclassT = TypeVar('_DataclassT')


class ArgumentParser(
    argparse_flags.ArgumentParser, simple_parsing.ArgumentParser
):
  """An `ArgumentParser` that handles both `simple_parsing` and `absl` flags.

  This class is a workaround for the fact that `simple_parsing.ArgumentParser`
  does not natively handle `absl.flags`. Without this, `absl` flags are not
  correctly parsed, especially when they are mixed with positional arguments,
  leading to errors.

  The `absl.flags.argparse_flags.ArgumentParser` is designed to integrate `absl`
  flags into an `argparse` setup. It does this by dynamically adding all
  defined `absl` flags to the parser instance upon initialization.

  By inheriting from both, we get the features of both:
  - `simple_parsing.ArgumentParser`: Allows defining arguments from typed
    dataclasses.
  - `argparse_flags.ArgumentParser`: Adds support for `absl` flags.

  The Method Resolution Order (MRO) is:
  `ArgumentParser` -> `argparse_flags.ArgumentParser` ->
  `simple_parsing.ArgumentParser` -> `argparse.ArgumentParser` -> `object`.

  This order is important. `argparse_flags.ArgumentParser` is first so that it
  can intercept arguments and handle `absl` flags before they are passed to
  `simple_parsing.ArgumentParser`.
  """

  def parse_known_args(
      self,
      args: Sequence[str] | None = None,
      namespace: argparse.Namespace | None = None,
      attempt_to_reorder: bool = False,
  ):
    # `argparse_flags.ArgumentParser` does not support `attempt_to_reorder` that
    # is used by `simple_parsing.ArgumentParser`. Since we don't need it, we can
    # just ignore it.
    del attempt_to_reorder
    if args:
      args = flag_utils.normalize_flags(args)
    return super().parse_known_args(args, namespace)


def make_flags_parser(
    args_dataclass: type[_DataclassT], description: str
) -> Callable[[list[str]], _DataclassT]:
  """Returns a function that parses flags and returns the dataclass instance."""

  def _parse_flags(argv: list[str]) -> _DataclassT:
    """Command lines flag parsing."""
    parser = ArgumentParser(
        description=description,
        allow_abbrev=False,
    )
    parser.add_arguments(args_dataclass, dest='args')
    return parser.parse_args(argv[1:]).args

  return _parse_flags


@dataclasses.dataclass(frozen=True, kw_only=True)
class Args(abc.ABC):
  """CLI arguments for TFDS CLI commands."""

  @abc.abstractmethod
  def execute(self) -> None:
    """Execute the CLI command."""
    ...


@dataclasses.dataclass
class DatasetInfo:
  """Structure for common string used for formatting.

  Attributes:
    name: Dataset name (`my_dataset`)
    cls_name: Class name (`MyDataset`)
    path: Root directory in which the dataset is added
    in_tfds: Whether the dataset is added in tensorflow_datasets/ or externally
    tfds_api: The Dataset API to import
    todo: Field to fill (`TODO(my_dataset)`)
    ds_import: Dataset import (`tensorflow_datasets.image.my_dataset`)
    data_format: Optional format of the input data used to generate
      format-specific dataset builders.
  """

  name: str
  in_tfds: bool
  path: pathlib.Path
  cls_name: str = dataclasses.field(init=False)
  tfds_api: str = dataclasses.field(init=False)
  todo: str = dataclasses.field(init=False)
  ds_import: str = dataclasses.field(init=False)
  data_format: str

  def __post_init__(self):
    self.cls_name = naming.snake_to_camelcase(self.name)
    self.tfds_api = (
        'tensorflow_datasets.public_api'
        if self.in_tfds
        else 'tensorflow_datasets'
    )
    self.todo = f'TODO({self.name.lower()})'

    if self.in_tfds:
      # `/path/to/tensorflow_datasets/image/my_dataset`
      # ->`tensorflow_datasets.image.my_dataset`
      import_parts = itertools.dropwhile(
          lambda p: p != 'tensorflow_datasets', self.path.parts
      )
      ds_import = '.'.join(import_parts)
    else:
      # For external datasets, it's difficult to correctly infer the full
      # `from my_module.path.datasets.my_dataset import MyDataset`.
      # Could try to auto-infer the absolute import path from the `setup.py`.
      # Instead uses relative import for now: `from . import my_dataset`
      ds_import = '.'
    self.ds_import = ds_import


@dataclasses.dataclass(frozen=True, kw_only=True)
class DebugOptions:
  """Debug & tests options.

  Attributes:
    overwrite: If True, delete pre-existing dataset if it exists.
    fail_if_exists: If True, fails the program if there is a pre-existing
      dataset.
    max_examples_per_split: When set, only generate the first X examples
      (default to 1), rather than the full dataset. If set to 0, only execute
      the `_split_generators` (which download the original data), but skip
      `_generator_examples`.
  """

  overwrite: bool = simple_parsing.flag(default=False)
  fail_if_exists: bool = simple_parsing.flag(default=False)
  max_examples_per_split: int | None = simple_parsing.field(
      default=None, nargs='?', const=1
  )


@dataclasses.dataclass(frozen=True, kw_only=True)
class PathOptions:
  """Path options.

  Attributes:
    data_dir: Where to place datasets. Defaults to `~/tensorflow_datasets/` or
      `TFDS_DATA_DIR` environement variable.
    download_dir: Where to place downloads. Defaults to `<data_dir>/downloads/`.
    extract_dir: Where to extract files. Defaults to
      `<download_dir>/extracted/`.
    manual_dir: Where to manually download data (required for some datasets).
      Defaults to `<download_dir>/manual/`.
    add_name_to_manual_dir: If true, append the dataset name to the `manual_dir`
      (e.g. `<download_dir>/manual/<dataset_name>/`). Useful to avoid collisions
      if many datasets are generated.
  """

  data_dir: epath.Path = simple_parsing.field(
      default=epath.Path(constants.get_default_data_dir())
  )
  download_dir: epath.Path | None = None
  extract_dir: epath.Path | None = None
  manual_dir: epath.Path | None = None
  add_name_to_manual_dir: bool = simple_parsing.flag(default=False)


@dataclasses.dataclass(frozen=True, kw_only=True)
class GenerationOptions:
  """Generation options.

  Attributes:
    download_only: If True, download all files but do not prepare the dataset.
      Uses the checksum.tsv to find out what to download. Therefore, this does
      not work in combination with --register_checksums.
    config: Config name to build. Build all configs if not set. Can also be a
      json of the kwargs forwarded to the config `__init__` (for custom
      configs).
    config_idx: Config id to build (`builder_cls.BUILDER_CONFIGS[config_idx]`).
      Mutually exclusive with `--config`. We are forced to have 2 flags to avoid
      ambiguity when `config` is a number (e.g. `voc/2017`).
    update_metadata_only: If True, existing dataset_info.json is updated with
      metadata defined in Builder class(es). Datasets must already have been
      prepared.
    download_config: A json of the kwargs forwarded to the config `__init__`
      (for custom DownloadConfigs).
    imports: Comma separated list of module to import to register datasets.
    register_checksums: If True, store size and checksum of downloaded files.
    force_checksums_validation: If True, raise an error if the checksums are not
      found. Otherwise, bypass the checks on the checksums
    beam_pipeline_options: A (comma-separated) list of flags to pass to
      `PipelineOptions` when preparing with Apache Beam. (see:
      https://www.tensorflow.org/datasets/beam_datasets). Example:
        `--beam_pipeline_options=job_name=my-job,project=my-project`
    file_format: File format to which generate the tf-examples.
    max_shard_size_mb: The max shard size in megabytes.
    num_shards: The number of shards to write to.
    num_processes: Number of parallel build processes.
    nondeterministic_order: If True, it will not assure deterministic ordering
      when writing examples to disk. This might result in quicker dataset
      preparation. Otherwise, it will assure deterministic ordering when writing
      examples to disk
  """

  download_only: bool = simple_parsing.flag(default=False)
  config: str | None = simple_parsing.field(default=None, alias='-c')
  config_idx: int | None = simple_parsing.field(
      default=None, alias='--builder_config_id'
  )
  update_metadata_only: bool = simple_parsing.flag(default=False)
  download_config: str | None = None
  imports: str | None = simple_parsing.field(
      default=None, alias=['-i', '--module_import']
  )
  register_checksums: bool = simple_parsing.flag(default=False)
  force_checksums_validation: bool = simple_parsing.flag(default=False)
  beam_pipeline_options: str | None = None
  file_format: str | None = simple_parsing.choice(
      *(file_format.value for file_format in file_adapters.FileFormat),
      default=None,
  )
  max_shard_size_mb: int | None = None
  num_shards: int | None = None
  num_processes: int = simple_parsing.field(default=1, alias='num-processes')
  nondeterministic_order: bool = simple_parsing.flag(default=False)


@dataclasses.dataclass(frozen=True, kw_only=True)
class PublishingOptions:
  """Publishing options.

  Attributes:
    publish_dir: Where to optionally publish the dataset after it has been
      generated successfully. Should be the root data dir under which datasets
      are stored. If unspecified, dataset will not be published.
    skip_if_published: If the dataset with the same version and config is
      already published, then it will not be regenerated.
  """

  publish_dir: epath.Path | None = None
  skip_if_published: bool = simple_parsing.flag(default=False)


@dataclasses.dataclass(frozen=True, kw_only=True)
class AutomationOptions:
  """Automation options.

  Attributes:
    exclude_datasets: If set, generate all datasets except the one defined here.
      Comma separated list of datasets to exclude.
    experimental_latest_version: Build the latest Version(experiments=...)
      available rather than default version.
  """

  exclude_datasets: str | None = None
  experimental_latest_version: bool = simple_parsing.flag(default=False)


def download_and_prepare(
    builder: dataset_builder.DatasetBuilder,
    download_config: download.DownloadConfig | None,
    download_dir: epath.Path | None,
    publish_dir: epath.Path | None,
    skip_if_published: bool,
    overwrite: bool,
    beam_pipeline_options: str | None,
    nondeterministic_order: bool = False,
) -> None:
  """Generate a single builder."""
  dataset = builder.info.full_name
  logging.info('download_and_prepare for dataset %s...', dataset)

  publish_data_dir = publish_dir / dataset if publish_dir else None
  if skip_if_published and publish_data_dir is not None:
    if publish_data_dir.exists():
      logging.info(
          'Dataset already exists in %s. Skipping generation.', publish_data_dir
      )
      return

  if not download_config:
    download_config = download.DownloadConfig()
  if overwrite and not download_config.download_mode.overwrite_dataset:
    download_config.download_mode = download.GenerateMode.REUSE_CACHE_IF_EXISTS
  if nondeterministic_order:
    download_config.nondeterministic_order = True

  # Add Apache Beam options to download config.
  try:
    import apache_beam as beam  # pylint: disable=g-import-not-at-top

    if beam_pipeline_options:
      download_config.beam_options = (
          beam.options.pipeline_options.PipelineOptions(
              flags=[f'--{opt}' for opt in beam_pipeline_options.split(',')]
          )
      )
  except ImportError:
    pass

  builder.download_and_prepare(
      download_dir=download_dir,
      download_config=download_config,
  )

  # Dataset generated successfully.
  logging.info('Dataset generation completed...')

  print()
  print(repr(builder.info))
  print()

  # Publish data if requested.
  if publish_data_dir is not None:
    logging.info('Publishing dataset to %s', publish_data_dir)
    file_utils.publish_data(
        from_data_dir=epath.Path(builder.data_dir),
        to_data_dir=publish_data_dir,
        overwrite=overwrite,
    )
    logging.info('Dataset successfully published to %s', publish_data_dir)
